"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Be able to search and replace editable regions within text files

.. py:data:: __all__
   :type: tuple[str, str]
   :value: ("Snip", "ReplaceResult")

   Modules exports

.. py:data:: _logger
   :type: logging.Logger

   Module level logger

"""

from __future__ import annotations

import logging
import re
import string
import sys
from enum import (
    Enum,
    auto,
)
from functools import partial
from pathlib import (
    Path,
    PurePath,
)

from .check_type import is_ok
from .constants import g_app_name

__package__ = "drain_swamp"
__all__ = (
    "Snip",
    "ReplaceResult",
)

_logger = logging.getLogger(f"{g_app_name}.snip")
is_module_debug = True


class ReplaceResult(Enum):
    """Snippet replace result possibilities

    For membership checking, equality comparison is supported

    .. code-block:: python

        oh_this_is_bad = ReplaceResult.VALIDATE_FAIL
        assert oh_this_is_bad == ReplaceResult.VALIDATE_FAIL

        hmm_looks_promising = ReplaceResult.REPLACED
        assert hmm_looks_promising != oh_this_is_bad

    .. py:attribute:: VALIDATE_FAIL

       Validation failed. Either nested or no matching start/end token

    .. py:attribute:: NO_MATCH

       No snippet with specified snippet code

    .. py:attribute:: REPLACED

       Snippet contents replaced

    .. py:attribute:: NO_CHANGE

       Snippet contents same as existing, so replace skipped

    """

    VALIDATE_FAIL = auto()
    NO_MATCH = auto()
    REPLACED = auto()
    NO_CHANGE = auto()

    def __eq__(self, other):
        """Equality check

        :param other: Should be same Enum class
        :type other: typing.Any
        :returns: True if equal otherwise False
        :rtype: bool
        """
        return self.__class__ is other.__class__ and other.value == self.value


def check_matching_tag_count(
    contents,
    token_start=None,
    token_end=None,
):
    """Checks tag count. Completely oblivious to overlapping tags

    Non-overlapping start/end tokens

    .. code-block:: python

        from drain_swamp.snip import check_matching_tag_count, Snip

        token_start = Snip.TOKEN_START
        token_end = Snip.TOKEN_END
        contents = f"{token_start}{token_end}{token_start}{token_end}"

        is_ok = check_matching_tag_count(
            contents,
            token_start,
            token_end,
        )
        assert is_ok

    :param contents: file contents to be validated
    :type contents: str
    :param token_start: Default None. Regex indicating start of a snippet
    :type token_start: str | None
    :param token_end: Default None. Regex indicating end of a snippet
    :type token_end: str | None
    :returns: True is check successful otherwise False
    :rtype: bool
    """
    if is_ok(contents) and is_ok(token_start) and is_ok(token_end):
        start_count = contents.count(token_start)
        end_count = contents.count(token_end)

        if is_module_debug:  # pragma: no cover
            _logger.debug(f"start token count: {start_count}")
            _logger.debug(f"end token count: {end_count}")
        else:  # pragma: no cover
            pass

        is_tag_count_match = start_count == end_count

        if is_module_debug:  # pragma: no cover
            _logger.debug(f"token counts match: {is_tag_count_match}")
        else:  # pragma: no cover
            pass

        ret = is_tag_count_match
    else:
        ret = False

    return ret


def check_not_nested_or_out_of_order(
    contents,
    token_start,
    token_end,
):
    """Check tag pairs are not nested / out of order

    Assumes :py:func:`drain_swamp.snip.check_matching_tag_count`
    already checked.

    :param contents: file contents to check
    :type contents: str
    :param token_start: Default None. Start token
    :type token_start: str | None
    :param token_end: Default None. End token
    :type token_end: str | None
    :returns:

       True on success False indicates give feedback and
       **definitely do not** attempt replacing a snippet

    :rtype: bool
    :raises:

        - :py:exc:`ValueError` -- Either no contents, no start token,
          or no end token provided

    """
    ret = True

    if is_ok(contents) and is_ok(token_start) and is_ok(token_end):
        idx_current = 0
        found_next = True
        while found_next is True:
            idx_start = contents.find(token_start, idx_current)
            idx_end = contents.find(token_end, idx_current)

            # Not errors:
            # - no tag pairs at all
            # - no next tags
            is_no_more_tags = idx_start == -1 and idx_end == -1
            if is_no_more_tags:
                found_next = False
                continue
            else:  # pragma: no cover
                pass

            # is_end_at_beginning = idx_end != -1 and idx_end == idx_current
            is_no_start_tag = idx_start == -1 and idx_end != -1
            is_no_end_tag = idx_start != -1 and idx_end == -1
            is_both_exist = idx_start != -1 and idx_end != -1
            is_end_before_start = is_both_exist and idx_end < idx_start
            is_bad = is_no_start_tag or is_no_end_tag or is_end_before_start

            # is_good
            """
            is_start_before_end = (
                is_both_exist
                and idx_start < idx_end
            )
            """
            pass

            if is_bad:
                ret = False
                found_next = False
            else:
                # next start idx
                idx_next_start = idx_end + len(token_end) + 1
                # advance past end tag
                idx_current = idx_next_start
    else:
        # either no contents, no start token, or no end token
        ret = False

        if is_module_debug:  # pragma: no cover
            msg_info = "Either no contents, no start token, or no end token provided"
            _logger.info(msg_info)
        else:  # pragma: no cover
            pass

    return ret


def sanitize_id(id_=""):
    """tokenize snippet code. alphanumeric + underscore

    :param id_:

       Snippet code. Default empty str. If empty str or None or just
       whitespace, considered as an empty string

    :type id_: typing.Any | None
    :returns: The snippet code sanitized
    :rtype: str
    """
    if id_ is None:
        # None
        ret = ""
    else:
        if not isinstance(id_, str):
            # unsupported type
            ret = ""
        else:
            # hyphen --> underscore
            ret = id_.replace("-", "_")

            # \w
            allowed_chars = string.ascii_letters + "_"
            # filter out any other character
            ret = "".join(c for c in ret if c in allowed_chars)

    return ret


class Snip:
    """jinja2 templates is time consuming to get the spacing right. A
    snippet is an easier alternative

    .. py:attribute:: TOKEN_START
       :type: str
       :value: "# @@@ editable"

       Beginning token of an editable section. Optionally can be followed
       by one whitespace and a str id. ID can contain alphanumberic
       characters and underscore. If there is no id, a file can contain at
       most one editable section

    .. py:attribute:: TOKEN_END
       :type: str
       :value: "# @@@ end\\n"

       End token denotes end of editable section. A trailing newline is expected

    .. py:attribute:: PATTERN_W_ID
       :type: re.Pattern

       Compiled regex which allows for an optional id. Captures two groups:
       id and contents.

       Removed regex which doesn't not support optional id for editable sections

       ``.*`` means capture greedily. ``.*?`` means shortest path match

       .. seealso::

          ``(?s)`` means single line mode. equivalent to compile flag ``re.DOTALL``

          `regex -- modifiers <https://www.regular-expressions.info/modifiers.html>`_

    :ivar fname: the file in/near the project to update
    :vartype fname: str | pathlib.Path
    :ivar is_quiet: Default False, set to True to repress logging
    :vartype is_quiet: bool | None

    :raises:

       - :py:exc:`TypeError` -- Unsupported type, fname can be either a str or Path

    .. todo::

       Take editable file paths from ``pyproject.toml``

    """

    TOKEN_START = "# @@@ editable"
    TOKEN_END = "# @@@ end\n"
    # PATTERN_WO_ID = re.compile(r"(?s)# @@@ editable\n.*# @@@ end\n")
    PATTERN_W_ID = re.compile(
        r"# @@@ editable\s?(\w*)?\n(.*?)\n# @@@ end\n",
        flags=re.DOTALL,
    )
    __slots__ = ("_is_quiet", "_path_file", "_contents", "_is_infer")

    def __init__(
        self,
        fname,
        is_quiet=False,
    ):
        super().__init__()
        self.is_quiet = is_quiet

        # Not throughly validated
        self.path_file = fname

        self._contents = None
        self._is_infer = False

    @property
    def path_file(self):
        """Path has not necessarily been checked. Run thru a validator func

        :returns: A raw path either relative or absolute
        :rtype: pathlib.Path
        """
        return self._path_file

    @path_file.setter
    def path_file(self, val):
        """Setter does not check:

        - whether relative or absolute path

        - whether is known to have editable region and therefore allowed

        :param val: Should be a str or Path. Either relative or absolute
        :type val: typing.Any
        :raises:

           - :py:exc:`TypeError` -- Unsupported type, can be either str or Path

        """
        msg_exc = "Unsupported type, fname can be either a str or Path"
        if val is None:
            raise TypeError(msg_exc)
        else:
            if isinstance(val, str):
                self._path_file = Path(val)
            elif issubclass(type(val), PurePath):
                self._path_file = val
            else:
                raise TypeError(msg_exc)

    def is_file_ok(self):
        """Try to grab permitted editable files from ``pyproject.toml``

        :returns: True if file has absolute and is permitted
        :rtype: bool

        .. todo:: r/w ??

           What about confirm: readable and writable!

        """
        # Replace this with a check of ``pyproject.toml`` settings
        is_editable_file = True
        path_f = self.path_file
        ret = (
            path_f.exists()
            and path_f.is_file()
            and path_f.is_absolute()
            and is_editable_file
        )
        return ret

    @property
    def is_quiet(self):
        """Whether to permit printing status to stderr

        :returns: True to not print status updates
        :rtype: bool
        """
        return self._is_quiet

    @is_quiet.setter
    def is_quiet(self, val):
        """Whether or not to suppress info logging messages

        :param val: Should be a bool. Defaults to False
        :type val: typing.Any
        """
        # Eventhough optional, is_quiet is a bool
        if val is None or not isinstance(val, bool):
            self._is_quiet = False
        else:  # pragma: no cover
            self._is_quiet = val

    @property
    def is_infer(self):
        """id is None or empty str. snippet_co is taken from snippet

        :returns: True if infer snippet_co otherwise False
        :rtype: bool
        """
        return self._is_infer

    def get_file(self):
        """Read the file

        :returns: file contents
        :rtype: str
        :raises:

           - :py:exc:`ValueError` -- cannot read file or is empty
           - :py:exc:`FileNotFoundError` -- file is not ok for whatever reason

        """
        if not self.is_file_ok() and not self.is_quiet:  # pragma: no cover
            msg_info = f"Cannot update nonexistent file, {str(self.path_file)}"
            _logger.info(msg_info)
        else:  # pragma: no cover
            pass

        if self.is_file_ok():
            ret = self.path_file.read_text(encoding="utf-8")

            if ret is None or len(ret) == 0:
                # Expecting file to not be empty
                # Also expecting to contain editable regions
                msg_exc = "Expecting file to be non-empty"
                raise ValueError(msg_exc)
            else:  # pragma: no cover
                pass
        else:  # pragma: no cover
            # file not ok
            msg_exc = (
                f"file not ok. replace editable region ... skip {str(self.path_file)}"
            )
            raise FileNotFoundError(msg_exc)

        return ret

    def replace(self, replacement, id_=""):
        """Find snippet with *id_*. If no *id_*, provided the file
        may contain at most one snippet.

        *id_* should be a cringe worthy cultural reference to an object
        or minor character! One id should be changed for every minor
        version release or PR

        In the changelog, in the commit, the top most lines should
        use

        .. code-block:: text

           *- style: * tags

           Explain the snippet_co (id_) cringe worthy cultural reference


        :param replacement: Just the content. Will be substituted within the file
        :type replacement: str
        :param id_:

           Default empty string. So as to support multiple snippets within a file.
           If only intended ever to have one snippet, empty string is appropriate

        :type id_: str | None
        :returns: VALIDATE_FAIL, NO_MATCH, REPLACED, NO_CHANGE
        :rtype: drain_swamp.snip.ReplaceResult

        :raises:

            - :py:exc:`TypeError` -- Unsupported type, *replacement* contents must be a str

        """
        cls = type(self)
        is_log_ok = not self.is_quiet
        if replacement is None or not isinstance(replacement, str):
            msg_exc = "Unsupported type, replacement contents must be a str"
            raise TypeError(msg_exc)

        id_ = sanitize_id(id_)
        if is_log_ok:  # pragma: no cover
            _logger.info(f"id_ (user input; filtered): {id_}")
            _logger.info(f"replacement (user input filtered): {replacement}")
        else:  # pragma: no cover
            pass

        def replace_fcn(matchobj):
            """inline func for snippets. Which are editable regions within utf-8 text
            files. Under the hood, The search and replace is using
            :py:func:`re.sub` and a nasty regex

            - ``matchobj.group(0)``
              Entire snippet including start/end tokens. Non-greedy
              will match first end token

            - ``matchobj.group(1)``
              snippet id. Can be empty string if no multiple id support

            - ``matchobj.group(2)``
              snippet contents

            Requires these outer func/meth variables:

            - token_start
            - token_end
            - id_
            - replacement

            Assumes the text is within a text file, not a stream nor a str.

            Example before and after

            snippet with an id

            .. code-block:: text

               id_ = "asdf"
               replacement = "abc abc abc"
               old_text = "zzzzzzzzzzzzz\n# @@@ editable asdf\nblah blah blah\n# @@@ end\nzzzzzzzzzzzz\n"

               expected = (
                   "zzzzzzzzzzzzz\n# @@@ editable asdf\nabc abc abc\n# @@@ end\nzzzzzzzzzzzz\n"
               )

            snippet without an id

            .. code-block:: text

               id_ = ""
               replacement = "abc abc abc"
               old_text = "zzzzzzzzzzzzz\n# @@@ editable\nblah blah blah\n# @@@ end\nzzzzzzzzzzzz\n"

               expected = (
                   "zzzzzzzzzzzzz\n# @@@ editable\nabc abc abc\n# @@@ end\nzzzzzzzzzzzz\n"
               )

            Only one snippet without an id allowed with a file. snippets
            with id and **one** snippet without id are allowed

            :param matchobj: Regex match object. Provided when a match occurs
            :type matchobj: re.Match
            :returns:

               If not an actual match, return the original content
               otherwise modify the content. Whatever returned will
               replace the existing text

            :rtype: str | None

            .. seealso::

               `re.sub <https://docs.python.org/3/library/re.html#re.sub>`_

               `regex tester web app <https://regexr.com/>`_

            """
            if is_log_ok:  # pragma: no cover
                _logger.debug(f"matched region:   {matchobj.group(0)}")
                _logger.debug(f"current id:       {matchobj.group(1)}")
                _logger.debug(f"current contents: {matchobj.group(2)}")
                _logger.debug(f"target id:        {id_}")
                _logger.debug(f"replacement: {replacement}")
            else:  # pragma: no cover
                pass

            current_id = matchobj.group(1)
            is_target_have_id = len(id_) != 0
            is_current_have_id = len(current_id) != 0

            if is_target_have_id and is_current_have_id and current_id == id_:
                # both have id and match
                ret = f"{token_start} {id_}\n{replacement}\n{token_end}"

                if is_log_ok:  # pragma: no cover
                    msg_debug = "both have id; match"
                    _logger.debug(msg_debug)
                else:  # pragma: no cover
                    pass
            elif is_target_have_id and is_current_have_id and current_id != id_:
                # both have id; no match
                if is_log_ok:  # pragma: no cover
                    msg_debug = "both have id; no match --> return unmodified"
                    _logger.debug(msg_debug)
                else:  # pragma: no cover
                    pass
                ret = matchobj.group(0)
            elif not is_target_have_id and not is_current_have_id:
                ret = f"{token_start}\n{replacement}\n{token_end}"
                if is_log_ok:  # pragma: no cover
                    msg_debug = "both no id; match"
                    _logger.debug(msg_debug)
                else:  # pragma: no cover
                    pass
            else:
                if is_log_ok:  # pragma: no cover
                    msg_debug = "not a match --> return unmodified"
                    _logger.debug(msg_debug)
                else:  # pragma: no cover
                    pass
                ret = matchobj.group(0)

            if is_log_ok:  # pragma: no cover
                msg_debug = f"ret: {ret}"
                _logger.debug(msg_debug)
            else:  # pragma: no cover
                pass

            return ret

        # first match snippet
        t_snippet_existing = self.contents(id_=id_)
        if isinstance(t_snippet_existing, ReplaceResult):
            # invalid file or one or more validation checks failed
            ret = t_snippet_existing
        else:
            # snippet_existing = t_snippet_existing[0]
            snippet_co_actual = t_snippet_existing[1]
            # If snippet_co not provided and only one snippet, use inferred snippet_co
            if self.is_infer:
                id_ = snippet_co_actual
            else:  # pragma: no cover
                pass

            # file contents
            text_existing = self._contents

            # Run re.sub on all editable regions
            token_start = cls.TOKEN_START
            token_end = cls.TOKEN_END
            new_text = re.sub(cls.PATTERN_W_ID, replace_fcn, text_existing)

            if is_log_ok:  # pragma: no cover
                msg_info = f"text (after re.sub): {new_text}"
                _logger.info(msg_info)
            else:  # pragma: no cover
                pass

            is_changed = new_text != text_existing

            if is_changed:
                if is_log_ok:  # pragma: no cover
                    msg_info = f"Updating {str(self.path_file)}"
                    _logger.info(msg_info)
                else:  # pragma: no cover
                    pass
                self.path_file.write_text(new_text)
                ret = ReplaceResult.REPLACED
            else:  # pragma: no cover
                ret = ReplaceResult.NO_CHANGE

        return ret

    def validate(self):
        """validate target file contents are safe

        All checks must pass. Over time add additional checks

        :returns: All checks passed
        :rtype: bool
        """

        # This is an initial check, so should return, not raise
        try:
            contents = self.get_file()
        except (ValueError, FileNotFoundError):
            msg_exc = (
                f"file not ok. replace editable region ... skip {str(self.path_file)}"
            )
            _logger.warning(msg_exc)
            self._contents = None
            return False

        cls = type(self)

        token_start = cls.TOKEN_START
        token_end = cls.TOKEN_END
        validators = (
            partial(
                check_matching_tag_count,
                contents,
                token_start=token_start,
                token_end=token_end,
            ),
            partial(
                check_not_nested_or_out_of_order,
                contents,
                token_start=token_start,
                token_end=token_end,
            ),
        )
        ret = all([validator() for validator in validators])

        # file contents --> buffer, so don't have to get_file twice
        if ret is True:
            self._contents = contents
        else:
            msg_exc = "Validation checks fail"
            _logger.warning(msg_exc)
            self._contents = None

        return ret

    @property
    def snippets(self):
        """Get all snippets. No filtering by snippet_co

        :returns: tuple of snippet_co and snippet contents
        :rtype: list[tuple[str, str]] | ReplaceResult
        """
        cls = type(self)

        is_log_ok = not self.is_quiet
        # If True --> self._contents is a str | None. None if checks fail
        is_valid = self.validate()
        is_invalid_file = not is_valid
        is_checks_fail = is_valid and self._contents is None
        if is_invalid_file or is_checks_fail:
            if is_log_ok:  # pragma: no cover
                msg_info = f"Validation issue. Check file: {str(self.path_file)}"
                _logger.info(msg_info)
            else:  # pragma: no cover
                pass

            return ReplaceResult.VALIDATE_FAIL

        # str, not None
        contents = self._contents

        pattern = cls.PATTERN_W_ID
        prog = re.compile(pattern)

        # snippet_co can be used multiple times, so not a set
        seq_ret = []
        for m in prog.finditer(contents):
            if m is not None:
                # id, snippet contents
                seq_ret.append((m.group(1), m.group(2)))
            else:  # pragma: no cover
                # no matches
                pass

        if len(seq_ret) == 0:
            return ReplaceResult.NO_MATCH

        return seq_ret

    def print(self):
        """Human readable summary of snippets

        :returns: tuple of snippet_co and snippet contents
        :rtype: list[tuple[str, str]] | ReplaceResult
        """
        snippets = self.snippets
        if (
            isinstance(snippets, ReplaceResult)
            and snippets == ReplaceResult.VALIDATE_FAIL
        ):
            msg = (
                "Snippet validation fail. Either nested or non-matching "
                "start/end tokens"
            )
            print(msg, file=sys.stderr)
        elif isinstance(snippets, ReplaceResult) and snippets == ReplaceResult.NO_MATCH:
            msg = "There are no snippets"
            print(msg, file=sys.stderr)
        else:
            lst = []
            lst_ids = []
            lst_contents = []
            for snippet_co, snippet_content in snippets:
                if len(snippet_co) == 0:
                    co = "(empty string)"
                else:
                    co = snippet_co
                lst_ids.append(co)
                block = f"{co} \n\n{snippet_content}"
                lst_contents.append(block)
            str_header_0 = "snippet codes:"
            str_header_1 = "blocks:"
            lst.append(f"{str_header_0}\n")
            lst.extend(lst_ids)
            lst.append(f"\n{str_header_1}\n")
            lst.extend(lst_contents)
            msg = "\n".join(lst)
            print(msg, file=sys.stderr)

        return snippets

    def contents(self, id_=None):
        """Get snippet contents.

        If only one snippet and id no provided, infer want the
        only available snippet.

        :param id_:

           Default None. snippet_co. If know there is only one snippet,
           my opt to not specify

        :type id_: str | None
        :returns:

           snippet contents and actual snippet_co. Possible to infer
           snippet if only one and id not provided

        :rtype: tuple[str, str] | drain_swamp.snip.ReplaceResult

        .. note:: context hint

           By providing a context hint, could achieve a better guess and
           have some awareness of context

        """
        id_ = sanitize_id(id_)

        snippets = self.snippets
        if isinstance(snippets, ReplaceResult):
            return snippets

        seq_ret = []
        # infer -- if only one and id not provided
        is_infer = len(snippets) == 1 and (
            id_ is None or (isinstance(id_, str) and len(id_.strip()) == 0)
        )
        if is_infer:
            # snippet_co not provided cuz knew there is only one snippet
            t_snip = snippets[0]
            snippet_co = t_snip[0]
            snippet_contents = t_snip[1]
            ret = (snippet_contents, snippet_co)
            self._is_infer = is_infer
        else:
            # matches
            for t_snip in snippets:
                snippet_co = t_snip[0]
                snippet_contents = t_snip[1]
                if snippet_co == id_:
                    seq_ret.append((snippet_contents, snippet_co))

            # ReplaceResult.NO_MATCH already checked for
            if len(seq_ret) == 0:
                ret = ReplaceResult.NO_MATCH
            else:
                ret = seq_ret[0]
            self._is_infer = False

        return ret



import re
from pathlib import Path
from typing import Final

__all__ = (
    "g_app_name",
    "package_name",
    "SUFFIX_IN",
    "SUFFIX_LOCKED",
    "SUFFIX_UNLOCKED",
    "SUFFIX_SYMLINK",
    "__version_app",
    "__url__",
    "PATH_PIP_COMPILE",
    "PROG_LOCK",
    "PROG_UNLOCK",
)

g_app_name: Final[str]
package_name: Final[str]
SUFFIX_IN: Final[str]
SUFFIX_LOCKED: Final[str]
SUFFIX_UNLOCKED: Final[str]
SUFFIX_SYMLINK: Final[str]

__version_app: Final[str]
__url__: Final[str]
_PATH_VENV: Path
PATH_PIP_COMPILE: Path
_pattern_lock: str
_pattern_unlock: str
PROG_LOCK: re.Pattern[str]
PROG_UNLOCK: re.Pattern[str]



__all__ = (
    "PyProjectTOMLParseError",
    "BackendNotSupportedError",
    "PyProjectTOMLReadError",
    "MissingRequirementsFoldersFiles",
)

class PyProjectTOMLReadError(OSError):
    def __init__(self, msg: str) -> None: ...

class PyProjectTOMLParseError(ValueError):
    def __init__(self, msg: str) -> None: ...

class BackendNotSupportedError(ValueError):
    def __init__(self, msg: str) -> None: ...

class MissingRequirementsFoldersFiles(AssertionError):
    def __init__(self, msg: str) -> None: ...



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. py:data:: __all__
   :type: tuple[str, str, str]
   :value: ("dump_version", "write_version_to_path", "write_version_files")

   Module exports

.. py:data:: TEMPLATES
   :type: dict[str, str]

   Templates to support both .py and .txt version files

.. seealso::

   Sources:

   - `[dump_version.py] <https://github.com/pypa/setuptools_scm/blob/main/src/setuptools_scm/dump_version.py>`_
   - `[_get_version_impl.py] <https://github.com/pypa/setuptools_scm/blob/main/src/setuptools_scm/_get_version_impl.py>`_

   `[LICENSE:MIT] <https://github.com/pypa/setuptools_scm/blob/main/LICENSE>`_

"""

import warnings
from pathlib import Path

from setuptools_scm._version_cls import _version_as_tuple

__all__ = (
    "dump_version",
    "write_version_to_path",
    "write_version_files",
)

TEMPLATES = {
    ".py": """\
# file generated by setuptools_scm
# don't change, don't track in version control
TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Tuple, Union
    VERSION_TUPLE = Tuple[Union[int, str], ...]
else:
    VERSION_TUPLE = object

version: str
__version__: str
__version_tuple__: VERSION_TUPLE
version_tuple: VERSION_TUPLE

__version__ = version = {version!r}
__version_tuple__ = version_tuple = {version_tuple!r}
""",
    ".txt": "{version}",
}


def dump_version(
    root,
    version,
    write_to,
    template=None,
) -> None:
    """Dump version file

    :param root: Absolute path of package base folder
    :type root: pathlib.Path
    :param version: Semantic version str
    :type version: str
    :param write_to: Should be relative path to version file
    :type write_to: pathlib.Path
    :param template: Default None. Template text
    :type template: str | None
    """
    assert isinstance(version, str)
    root = Path(root)
    write_to = Path(write_to)
    if write_to.is_absolute():
        # trigger warning on escape
        write_to.relative_to(root)
        warnings.warn(
            f"{write_to=!s} is a absolute path,"
            " please switch to using a relative version file",
            DeprecationWarning,
        )
        target = write_to
    else:
        target = Path(root).joinpath(write_to)

    write_version_to_path(target, template=template, version=version)


def _validate_template(
    target,
    template,
):
    """Provide a template or use built-in templates for .py or .txt target files

    :param target: absolute path to target file
    :type target: pathlib.Path
    :param template: Default None. Can pass in a custom template
    :type template: str | None
    :returns: template for _version file. Either ``.txt`` or ``.py``
    :rtype: str
    :meta private:
    :raises:

       - :py:exc:`ValueError` -- unsupported template file format.
         Support only ``*.txt`` and ``*.py``

    """
    is_empty = (
        template is not None
        and isinstance(template, str)
        and len(template.strip()) == 0
    )
    if is_empty:
        msg_warn = f"{template=} looks like a error, using default instead"
        warnings.warn(msg_warn)
        template = None
    else:  # pragma: no cover
        pass

    if template is None:
        template = TEMPLATES.get(target.suffix)
    else:
        pass

    if template is None:
        msg_exc = (
            f"bad file format: {target.suffix!r} (of {target})\n"
            "only *.txt and *.py have a default template"
        )
        raise ValueError(msg_exc)
    else:  # pragma: no cover
        pass

    return template


def write_version_to_path(
    target,
    template,
    version,
):
    """Write the version file

    :param target: Absolute path to _version.[suffix] file
    :type target: pathlib.Path
    :param template: Template str. Format fields version and version tuple
    :type template: str | None
    :param version: Semantic version str
    :type version: str
    :raises:

       - :py:exc:`ValueError` -- unsupported template file format.
         Support only ``*.txt`` and ``*.py``

    """
    try:
        final_template = _validate_template(target, template)
    except ValueError:
        raise
    # log.debug("dump %s into %s", version, target)

    # t_ver = SemVersion.as_tuple(version)
    t_ver = _version_as_tuple(version)

    content = final_template.format(version=version, version_tuple=t_ver)

    target.write_text(content, encoding="utf-8")


def write_version_files(
    version,
    root,
    write_to,
    version_file,
    is_only_not_exists=False,
):
    """Write the _version.py file

    Disabled passing in a custom template

    :param version: Semantic version str
    :type version: str
    :param root: package base folder Path
    :type root: pathlib.Path
    :param write_to: Testing target relative path
    :type write_to: str | None
    :param version_file: target relative path from pyproject.toml
    :type version_file: str | None
    :param is_only_not_exists: Default False. Write file only if it does not already exit
    :type is_only_not_exists: bool | None
    :raises:

       - :py:exc:`ValueError` -- unsupported template file format.
         Support only ``*.txt`` and ``*.py``

    .. note::

       Moved from setuptools_scm._get_version_impl

    """
    if is_only_not_exists is None or not isinstance(is_only_not_exists, bool):
        is_only_not_exists = False
    else:  # pragma: no cover
        pass

    def is_do_it(abs_path):
        nonlocal is_only_not_exists
        path_abs = Path(abs_path)
        if is_only_not_exists is False:
            ret = True
        else:
            if not Path(path_abs).exists():
                ret = True
            else:
                ret = False

        return ret

    if write_to is not None:
        # For testing?!
        is_run = is_do_it(write_to)
        if is_run:
            try:
                dump_version(
                    root=root,
                    version=version,
                    write_to=write_to,
                    template=None,
                )
            except ValueError:
                raise
        else:  # pragma: no cover
            pass
    else:  # pragma: no cover
        pass

    if version_file:
        # relative path from pyproject.toml
        version_file = Path(version_file)
        assert not version_file.is_absolute(), f"{version_file=}"
        target = root.joinpath(version_file)
        is_run = is_do_it(target)
        if is_run:
            try:
                write_version_to_path(
                    target,
                    template=None,
                    version=version,
                )
            except ValueError:  # pragma: no cover
                raise
        else:  # pragma: no cover
            pass
    else:  # pragma: no cover
        pass



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Without ``tool.pipenv-unlock.folders``:

   The folders containing .in files is deduced from
   ``[tool.pipenv-unlock]`` ``required`` and ``optionals`` fields. Which contains
   keys: ``target`` and ``relative_path``.

   ``relative_path`` value contains the relative path to a .in file.
   Without dups, use a set, those files' parent folder will contain our .in files.

With ``tool.pipenv-unlock.folders``:

There may be additional folders not implied by required/optionals
(aka ``dependencies`` or ``optional-dependencies``), which contain
``.in`` files.

Example use cases

-  ``ci``
   Used by CI/CD. e.g. mypy.in, tox.in

- ``kit``
  For building tarball and wheels, e.g. kit.in

In which case, there needs to be a way to specify all folders
containing ``.in`` files.

Explicitly specifying all folders is preferred over a derived
(implied) folders list

Example ``pyproject.toml``. specifies an additional folder, ``ci``.

.. code-block:: text

   [tool.pipenv-unlock]
   folders = [
       "docs",
       "requirements",
       "ci",
   ]
   required = { target = "prod", relative_path = "requirements/prod.in" }
   optionals = [
       { target = "pip", relative_path = "requirements/pip.in" },
       { target = "pip_tools", relative_path = "requirements/pip-tools.in" },
       { target = "dev", relative_path = "requirements/dev.in" },
       { target = "manage", relative_path = "requirements/manage.in" },
       { target = "docs", relative_path = "docs/requirements.in" },
   ]

.. py:data:: __all__
   :type: tuple[str, str, str]
   :value: ("lock_compile", "refresh_links", "unlock_compile")

   Module exports

.. py:data:: _logger
   :type: logging.Logger

   Module level logger

.. py:data:: is_module_debug
   :type: bool
   :value: False

   on/off for module level logging

"""

from __future__ import annotations

import copy
import dataclasses
import fileinput
import importlib.util
import logging
import os
import pathlib
from collections.abc import Sequence
from pathlib import (
    Path,
    PurePath,
)

from ._run_cmd import run_cmd
from .constants import (
    PATH_PIP_COMPILE,
    SUFFIX_LOCKED,
    SUFFIX_SYMLINK,
    SUFFIX_UNLOCKED,
    g_app_name,
)
from .exceptions import MissingRequirementsFoldersFiles

__package__ = "drain_swamp"
__all__ = (
    "lock_compile",
    "refresh_links",
    "unlock_compile",
)

_logger = logging.getLogger(f"{g_app_name}.lock_toggle")
is_module_debug = False


def is_piptools():
    """Check whether package pip-tools is installed. If not the
    :command:`pip-compile` would not be available

    This function is patchable

    :returns: True if pip-tools installed otherwise False
    :rtype: bool
    """
    spec = importlib.util.find_spec("piptools")
    is_package = spec is not None
    return is_package


def _create_symlinks_relative(src: str, dest: str, cwd_path: str) -> None:
    """Create the relative symlink

    src and dest are relative paths. Relative to cwd_path

    - No :py:func:`os.chdir`

    - Does not create the folder

    - Does not create the source file

    :param src: relative path to source file
    :type src: str
    :param dest: relative path to what will be the symlink
    :type dest: str
    :param cwd_path: Absolute path a folder
    :type cwd_path: str

    :raises:

       - :py:exc:`NotADirectoryError` -- Directory not found or not a directory
       - :py:exc:`FileNotFoundError` -- Source file not found
       - :py:exc:`NotImplementedError` -- os.symlink not supported on this platform
       - :py:exc:`ValueError` -- destination symlink must suffix must be .lnk

    :meta private:

    .. note::

       :code:`tempfile.TemporaryDirectory(delete=False)` delete keyword is py312+

    """
    path_cwd = Path(cwd_path)
    path_src = path_cwd.joinpath(src)
    path_parent = path_src.parent
    src_name = path_src.name
    dest_name = Path(dest).name
    path_parent_src = path_parent.joinpath(src_name)
    path_parent_dest = path_parent.joinpath(dest_name)

    # base folder, not including subfolders
    is_dir_bad = not path_cwd.exists() or (path_cwd.exists() and not path_cwd.is_dir())
    if is_dir_bad:
        msg_exc = "Expecting folder to already exist"
        raise NotADirectoryError(msg_exc)

    is_not_src = not path_parent_src.exists()
    if is_not_src:
        msg_exc = f"Source file not found: {src} folder: {path_cwd}"
        raise FileNotFoundError(msg_exc)

    # Assert dest suffixes indicate it's a .lnk file
    if path_parent_dest.suffixes != [
        SUFFIX_SYMLINK,
    ]:
        msg_exc = "Destination symlink must suffix must be .lnk"
        raise ValueError(msg_exc)

    # Remove dest (symlink), if it exists
    is_dest = path_parent_dest.exists() and (
        path_parent_dest.is_file() or path_parent_dest.is_symlink()
    )
    if is_dest:  # pragma: no cover
        path_parent_dest.unlink(missing_ok=True)
    else:  # pragma: no cover
        pass

    # Create file descriptor for tmp folder
    # folder is rw
    fd = os.open(str(path_parent), os.O_RDONLY)

    mod_path = f"{g_app_name}.lock_toggle._create_symlinks_relative"
    if is_module_debug:  # pragma: no cover
        _logger.info(f"{mod_path} parent {path_parent!r}")
        _logger.info(f"{mod_path} src name {src_name!r}")
        _logger.info(f"{mod_path} dest name {dest_name!r}")
    else:  # pragma: no cover
        pass

    try:
        # Create relative symlink
        os.symlink(src_name, dest_name, dir_fd=fd)
    except NotImplementedError:
        raise


def _maintain_symlink(path_cwd, abspath_src):
    """Create/Update symlink which indicates current state file

    ``%.lnk`` --> [``%.unlock`` | ``%.lock`]

    Symlink path must be relative, not absolute to the machine base folder
    :param path_cwd:

       folder containing the src file. Relative to the package or a requirements folder

    :type path_cwd: pathlib.Path
    :param abspath_src: Absolute path to the source file. With suffix .unlock or .lock
    :type abspath_src: pathlib.Path
    :raises:

       - :py:exc:`NotADirectoryError` -- Directory not found or not a directory
       - :py:exc:`FileNotFoundError` -- Source file not found
       - :py:exc:`NotImplementedError` -- os.symlink not supported on this platform

    :meta private:
    """
    is_dir_bad = not path_cwd.exists() or (path_cwd.exists() and not path_cwd.is_dir())
    if is_dir_bad:
        msg_exc = "Expecting folder to already exist"
        raise NotADirectoryError(msg_exc)

    is_not_src = not abspath_src.exists()
    if is_not_src:
        msg_exc = f"Source file not found: {abspath_src} folder: {path_cwd}"
        raise FileNotFoundError(msg_exc)

    relpath_src = abspath_src.relative_to(path_cwd)
    src = str(relpath_src)

    # dest suffix is .lnk, so ValueError impossible
    src_stem = abspath_src.stem
    dest = f"{src_stem}{SUFFIX_SYMLINK}"

    try:
        _create_symlinks_relative(src, dest, str(path_cwd))
    except Exception:  # pragma: no cover
        raise


def _postprocess_abspath_to_relpath(path_out, path_parent):
    """Within a lock file (contents), if an absolute path make relative
    by removing parent path

    To see the lock file format

    .. code-block:: shell

       pip-compile --dry-run docs/requirements.in

    :param path_out: Absolute path of the requirements file
    :type path_out: pathlib.Path
    :param path_parent: Absolute path to the parent folder of the requirements file
    :type path_parent: pathlib.Path
    """
    files = (path_out,)
    # py310 encoding="utf-8"
    with fileinput.input(files, inplace=True) as f:
        for line in f:
            is_lock_requirement_line = line.startswith("    # ")
            if is_lock_requirement_line:
                # process line
                line = line.replace(f"{path_parent!s}/", "")
            else:  # pragma: no cover
                # do not modify line
                pass


def lock_compile(inst):
    """In a subprocess, call :command:`pip-compile` to create ``.lock`` files

    :param inst:

       Backend subclass instance which has folders property containing
       ``collections.abc.Sequence[Path]``

    :type inst: drain_swamp.backend_abc.BackendType
    :returns: Generator of abs path to ``.lock`` files
    :rtype: collections.abc.Generator[pathlib.Path, None, None]
    :raises:

       - :py:exc:`AssertionError` -- pip-tools is not installed and is
         a dependency of this package

    """
    str_func_name = f"{g_app_name}.lock_toggle.lock_compile"
    assert is_piptools()

    # store pairs
    lst_pairs = []
    t_excludes = ("pins.in",)

    # Look at the folders. Then convert all ``.in`` --> ``.lock``
    gen_unlocked_files = inst.in_files()

    in_files = list(gen_unlocked_files)
    del gen_unlocked_files

    if is_module_debug:  # pragma: no cover
        msg_info = f"{str_func_name} in_files: {in_files}"
        _logger.info(msg_info)
    else:  # pragma: no cover
        pass

    gen_unlocked_files = inst.in_files()
    for path_abs in gen_unlocked_files:
        if path_abs.name not in t_excludes:
            file_name = f"{path_abs.stem}{SUFFIX_LOCKED}"
            abspath_locked = path_abs.parent.joinpath(file_name)
            lst_pairs.append((str(path_abs), str(abspath_locked)))
        else:  # pragma: no cover
            # Not a ``*.in`` file. Ignore
            pass

    if is_module_debug:  # pragma: no cover
        msg_info = f"{str_func_name} pairs: {lst_pairs}"
        _logger.info(msg_info)
    else:  # pragma: no cover
        pass

    # Serial it's whats for breakfast
    for in_path, out_path in lst_pairs:
        cmd = (
            str(PATH_PIP_COMPILE),
            "--allow-unsafe",
            "--resolver",
            "backtracking",
            "-o",
            out_path,
            in_path,
        )

        if is_module_debug:  # pragma: no cover
            msg_info = f"{str_func_name} cmd: {cmd}"
            _logger.info(msg_info)
        else:  # pragma: no cover
            pass

        run_cmd(cmd, cwd=inst.parent_dir)
        path_out = Path(out_path)
        is_confirm = path_out.exists() and path_out.is_file()
        if is_confirm:
            if is_module_debug:  # pragma: no cover
                msg_info = f"{str_func_name} yield: {out_path!s}"
                _logger.info(msg_info)
            else:  # pragma: no cover
                pass

            # post processing
            _postprocess_abspath_to_relpath(path_out, inst.parent_dir)

            yield path_out
        else:
            # File not created. Darn you pip-compile!
            yield from ()

    yield from ()


def strip_inline_comments(val):
    """Strip off inline comments. Which may be to the right of a requirement

    :param val: line with contains a requirement and optionally an in-line comment
    :type val: str
    :returns: Requirement without a inline comment
    :rtype: str
    """
    try:
        pos = val.index("#")
    except ValueError:
        # not found
        ret = val
    else:
        ret = val[:pos]
        ret = ret.rstrip()

    return ret


@dataclasses.dataclass
class InFile:
    """
    :ivar relpath: Relative path to requirements file
    :vartype relpath: str
    :ivar stem:

       Requirements file stem. Later, appends suffix ``.unlock``

    :vartype stem: str
    :ivar constraints:

       Requirement files may contain lines starting with
       ``-c [requirements file relative path]``. This constitutes a
       constraint. The requirements file referenced by a constraint, can
       also contain constraints. The tree of constraints is resolved
       recursively until all constraints on all requirements files are resolved.

    :vartype constraints: set[str]
    :ivar requirements:

       Contains all dependencies from a requirements file. There is no
       attempt made to resolve package versions.

    :vartype requirements: set[str]
    """

    relpath: str
    stem: str
    constraints: set[str] = dataclasses.field(default_factory=set)
    requirements: set[str] = dataclasses.field(default_factory=set)

    def __post_init__(self):
        """relpath given as a Path, convert into a str.
        :py:func:`drain_swamp.lock_toggle.InFile.check_path` should
        have already been performed/called prior
        """
        is_path = issubclass(type(self.relpath), PurePath)
        if is_path:
            self.relpath = str(self.relpath)
        else:  # pragma: no cover
            pass

    @staticmethod
    def check_path(cwd, path_to_check) -> None:
        """Check Path. Should not be a str

        :param cwd: Package base folder
        :type cwd: pathlib.Path
        :param path_to_check: Hopefully a relative Path
        :type path_to_check: typing.Any
        :raises:

           - :py:exc:`TypeError` -- Sequence contains one or more unsupported types
           - :py:exc:`ValueError` -- Requirements file, (.in), not relative to base folder
           - :py:exc:`FileNotFoundError` -- Requirements file, (.in), not found

        """
        # contains only Path
        is_path = path_to_check is not None and issubclass(
            type(path_to_check), PurePath
        )
        if not is_path:
            msg_exc = (
                f"in_files Sequence contains unsupported type, {type(path_to_check)}"
            )
            raise TypeError(msg_exc)
        else:  # pragma: no cover
            pass

        # FileNotFoundError
        is_abs_path = path_to_check.is_absolute() and path_to_check.is_file()
        if not is_abs_path:
            msg_exc = "Requirement (.in) file does not exist"
            raise FileNotFoundError(msg_exc)

        # relative to self.cwd
        try:
            path_to_check.relative_to(cwd)
        except Exception as e:
            msg_exc = (
                f"requirements file, {path_to_check}, not relative to folder, {cwd}"
            )
            raise ValueError(msg_exc) from e

    def abspath(self, path_package_base):
        """Get the absolute path. The relative path is relative to the
        package folder.

        :param path_package_base: package base folder
        :type path_package_base: pathlib.Path
        :returns: absolute path
        :rtype: pathlib.Path
        """
        return path_package_base.joinpath(self.relpath)

    @property
    def depth(self):
        """Number of unresolved constraints. One this number gets down
        to zero, the InFile is moved from files set --> zeroes set

        :returns: unresolved constraints count
        :rtype: int
        """
        return len(self.constraints)

    def resolve(self, constraint, requirements):
        """
        :param constraint: A ``.in`` file relative path
        :type constraint: str
        :param requirements:

           The ``.in`` file's requirement lines, which might have silly
           version upper limits. No attempt is made to address these
           upper bounds version limits

        :type requirements: set[str]
        """
        self.constraints.remove(constraint)

        # Removes duplicates, but ignores version constraints
        for req in requirements:
            self.requirements.add(req)

    def __hash__(self):
        """Constraints as constraints are resolved, are removed,
        increasing the requirements.

        Both fields are dynamic. For the point of identification,
        the relpath is unique

        :returns: hash of relpath
        :rtype: int
        """
        return hash((self.relpath,))

    def __eq__(self, right):
        """Compares equality

        :param right: right side of the equal comparison
        :type right: typing.Any
        :returns:

           True if both are same InFile otherwise False. Does not take
           in account, constraints and requirements.

        :rtype: bool
        """
        is_infile = isinstance(right, InFile)
        is_str = isinstance(right, str)
        is_relpath = issubclass(type(right), PurePath) and not right.is_absolute()
        if is_relpath:
            str_right = str(right)
        elif is_str:
            str_right = right
        else:  # pragma: no cover
            pass

        if is_infile:
            is_eq = self.__hash__() == right.__hash__()
            ret = is_eq
        elif is_str or is_relpath:
            # relpath
            left_hash = hash(self)
            right_hash = hash((str_right,))
            is_eq = left_hash == right_hash
            ret = is_eq
        else:
            ret = False

        return ret


@dataclasses.dataclass
class InFiles:
    """Container of InFile

    :ivar cwd: current working directory
    :vartype cwd: pathlib.Path
    :ivar in_files: Requirements files. Relative path to ``.in`` files
    :vartype in_files: collections.abc.Sequence[pathlib.Path]
    :ivar _files:

       Set of InFile. Which contains the relative path to a Requirement
       file. May contain unresolved constraints

    :vartype _files: set[InFile]
    :ivar _zeroes: Set of InFile that have all constraints resolved
    :vartype _zeroes: set[InFile]

    :raises:

       - :py:exc:`TypeError` -- in_files unsupported type, expecting
         ``Sequence[Path]``

       - :py:exc:`ValueError` -- An element within in_files is not
         relative to folder, cwd

       - :py:exc:`FileNotFoundError` -- Requirements .in file not found

       - :py:exc:`drain_swamp.exceptions.MissingRequirementsFoldersFiles` --
         A requirements file references a nonexistent constraint

    """

    cwd: pathlib.Path
    in_files: dataclasses.InitVar[Sequence[pathlib.Path]]
    _files: set[InFile] = dataclasses.field(init=False, default_factory=set)
    _zeroes: set[InFile] = dataclasses.field(init=False, default_factory=set)

    def __post_init__(self, in_files):
        """Read in and initial pass over ``.in`` files

        :param in_files: Requirements files. Relative path to ``.in`` files
        :type in_files: collections.abc.Sequence[pathlib.Path]
        """
        # is a sequence
        if in_files is None or not isinstance(in_files, Sequence):
            msg_exc = f"Expecting a list[Path] got unsupported type {in_files}"
            raise TypeError(msg_exc)

        # Checks
        for path_abs in in_files:
            try:
                InFile.check_path(self.cwd, path_abs)
            except (TypeError, ValueError, FileNotFoundError):
                raise

        for path_abs in in_files:
            path_relpath = path_abs.relative_to(self.cwd)
            str_file = path_abs.read_text()
            lines = str_file.split("\n")
            constraint_raw = []
            requirement = set()
            for line in lines:
                is_comment = line.startswith("#")
                is_blank_line = len(line.strip()) == 0
                is_constraint = line.startswith("-c ")
                if is_comment or is_blank_line:
                    continue
                elif is_constraint:
                    line_pkg = line[3:]
                    line_pkg = strip_inline_comments(line_pkg)
                    constraint_raw.append(line_pkg)
                else:
                    line_pkg = strip_inline_comments(line)
                    requirement.add(line_pkg)

            """Normalize constraint
            Assume .in files constraints are relative path only
            """
            path_parent = path_abs.parent
            constraint = set()
            for cons_path in constraint_raw:
                try:
                    path_abs_constraint = path_parent.joinpath(cons_path).resolve(
                        strict=True
                    )
                except FileNotFoundError as exc:
                    msg_exc = (
                        f"Within requirements file, {path_relpath}, a constraint "
                        f"file does not exist. Create it! {cons_path}"
                    )
                    raise MissingRequirementsFoldersFiles(msg_exc) from exc
                else:
                    path_rel_constraint = path_abs_constraint.relative_to(self.cwd)
                    constraint.add(str(path_rel_constraint))

            # Checks already performed for: TypeError, ValueError or FileNotFoundError
            in_ = InFile(
                relpath=path_relpath,
                stem=path_abs.stem,
                constraints=constraint,
                requirements=requirement,
            )
            if is_module_debug:  # pragma: no cover
                _logger.info(f"in_: {repr(in_)}")
            else:  # pragma: no cover
                pass
            # set.add an InFile
            self.files = in_

    @property
    def files(self):
        """Generator of InFile

        :returns: Yields InFile. These tend to contain constraints
        :rtype: collections.abc.Generator[drain_swamp.lock_toggle.InFile, None, None]
        """
        yield from self._files

    @files.setter
    def files(self, val):
        """append an InFile

        The constructor performs checks. Not intended to add requirements
        files outside of the constructor

        :param val: Supposed to be an :py:class:`~drain_swamp.lock_toggle.InFile`
        :type val: typing.Any
        """
        is_infile = val is not None and isinstance(val, InFile)
        if is_infile:
            self._files.add(val)
        else:  # pragma: no cover
            pass

    @property
    def zeroes(self):
        """Generator of InFile

        :returns: Yields InFile without any constraints
        :rtype: collections.abc.Generator[drain_swamp.lock_toggle.InFile, None, None]
        """
        yield from self._zeroes

    @zeroes.setter
    def zeroes(self, val):
        """append an InFile that doesn't have any constraints

        The only acceptable source of zeroes is from :code:`self._files`

        :param val: Supposed to be an :py:class:`~drain_swamp.lock_toggle.InFile`
        :type val: typing.Any
        """
        is_infile = val is not None and isinstance(val, InFile)
        if is_infile:
            self._zeroes.add(val)
        else:  # pragma: no cover
            pass

    def in_generic(self, val, set_name="files"):
        """A generic __contains__

        :param val: item to check if within zeroes
        :type val: typing.Any
        :param set_name: Default "files". Which set to search thru. zeroes or files.
        :type set_name: str | None
        :returns: True if InFile contained within zeroes otherwise False
        :rtype: bool
        """
        if set_name is None or not isinstance(set_name, str):
            set_name = "_files"
        else:  # pragma: no cover
            set_name = f"_{set_name}"

        ret = False
        set_ = getattr(self, set_name, set())
        for in_ in set_:
            if val is not None:
                is_match_infile = isinstance(val, InFile) and in_ == val
                is_match_str = isinstance(val, str) and in_.relpath == val

                is_match_path = issubclass(type(val), PurePath) and in_.relpath == str(
                    val
                )
                if is_match_infile or is_match_str or is_match_path:
                    ret = True
                else:  # pragma: no cover
                    # unsupported type
                    pass
            else:  # pragma: no cover
                # is None
                pass

        return ret

    def in_zeroes(self, val):
        """Check if within zeroes

        :param val: item to check if within zeroes
        :type val: typing.Any
        :returns: True if InFile contained within zeroes otherwise False
        :rtype: bool
        """

        return self.in_generic(val, set_name="zeroes")

    def __contains__(self, val):
        """Check if within InFiles

        :param val: item to check if within InFiles
        :type val: typing.Any
        :returns: True if InFile contained within InFiles otherwise False
        :rtype: bool
        """
        return self.in_generic(val)

    def get_by_relpath(self, relpath, set_name="files"):
        """Get the index and :py:class:`~drain_swamp.lock_toggle.InFile`

        :param relpath: relative path of a ``.in`` file
        :type relpath: str
        :param set_name: Default "files". Which set to search thru. zeroes or files.
        :type set_name: str | None
        :returns:

           The ``.in`` file and index within
           :py:class:`~drain_swamp.lock_toggle.InFiles`

        :rtype: drain_swamp.lock_toggle.InFile | None
        :raises:

            - :py:exc:`ValueError` -- Unsupported type. relpath is neither str nor Path

        """
        if set_name is None or not isinstance(set_name, str):
            set_name = "_files"
        else:  # pragma: no cover
            set_name = f"_{set_name}"

        msg_exc = f"Expected a relative path as str or Path. Got {type(relpath)}"
        str_relpath = None
        if relpath is not None:
            if isinstance(relpath, str):
                str_relpath = relpath
            elif issubclass(type(relpath), PurePath):
                str_relpath = str(relpath)
            else:
                raise ValueError(msg_exc)
        else:
            raise ValueError(msg_exc)

        ret = None
        set_ = getattr(self, set_name, set())
        for in_ in set_:
            if in_.relpath == str_relpath:
                ret = in_
                break
            else:  # pragma: no cover
                # not a match
                pass
        else:
            # set empty
            ret = None

        return ret

    def move_zeroes(self):
        """Zeroes have had all their constraints resolved and therefore
        do not need to be further scrutinized.
        """
        # add to self.zeroes
        del_these = []
        for in_ in self.files:
            if in_.depth == 0:
                # set.add an InFile
                self.zeroes = in_
                del_these.append(in_)
            else:  # pragma: no cover
                pass

        if is_module_debug:  # pragma: no cover
            msg_info = f"self.zeroes (after): {self._zeroes}"
            _logger.info(msg_info)
        else:  # pragma: no cover
            pass

        # remove from self._files
        for in_ in del_these:
            self._files.remove(in_)

        if is_module_debug:  # pragma: no cover
            msg_info = f"self.files (after): {self._files}"
            _logger.info(msg_info)
        else:  # pragma: no cover
            pass

    def resolve_zeroes(self):
        """If a requirements file have constraint(s) that can be
        resolved, by a zero, do so.

        _files and _zeroes are both type, set. Modifying an element
        modifies element within the set
        """
        # Take the win, early and often!
        self.move_zeroes()

        # Resolve with zeroes
        for in_ in self.files:
            constaints_copy = copy.deepcopy(in_.constraints)
            for constraint_relpath in constaints_copy:
                is_in_zeroes = self.in_zeroes(constraint_relpath)

                if is_module_debug:  # pragma: no cover
                    msg_info = (
                        f"resolve_zeroes constraint {constraint_relpath} in "
                        f"zeroes {is_in_zeroes}"
                    )
                    _logger.info(msg_info)
                else:  # pragma: no cover
                    pass

                if is_in_zeroes:
                    # Raises ValueError if constraint_relpath is neither str nor Path
                    item = self.get_by_relpath(constraint_relpath, set_name="zeroes")

                    if is_module_debug:  # pragma: no cover
                        msg_info = f"resolve_zeroes in_ (before) {in_}"
                        _logger.info(msg_info)
                    else:  # pragma: no cover
                        pass

                    in_.resolve(constraint_relpath, item.requirements)

                    if is_module_debug:  # pragma: no cover
                        msg_info = f"resolve_zeroes in_ (after) {in_}"
                        _logger.info(msg_info)
                    else:  # pragma: no cover
                        pass
                else:  # pragma: no cover
                    pass

        # For an InFile, are all it's constraints resolved?
        self.move_zeroes()

    def resolution_loop(self):
        """Run loop of resolve_zeroes calls, sampling before and after
        counts. If not fully resolved and two iterations have the same
        result, raise an Exception

        :raises:

           - :py:exc:`drain_swamp.exceptions.MissingRequirementsFoldersFiles` --
             there are unresolvable constraint(s)

        """
        initial_count = len(list(self.files))
        current_count = initial_count
        previous_count = initial_count
        while current_count != 0:
            self.resolve_zeroes()
            current_count = len(list(self.files))
            # Check previous run results vs current run results, if same raise Exception
            is_resolved = current_count == 0
            is_same_result = previous_count == current_count

            # raise exception if not making any progress
            if not is_resolved and is_same_result:
                unresolvable_requirement_files = [in_.relpath for in_ in self.files]
                missing_contraints = [in_.constraints for in_ in self.files]
                msg_exc = (
                    "Missing .in requirements file(s). Unable to resolve "
                    "constraint(s). Files with unresolvable constraints: "
                    f"{unresolvable_requirement_files}. "
                    f"Missing constraints: {missing_contraints}"
                )
                _logger.info(msg_exc)
                raise MissingRequirementsFoldersFiles(msg_exc)
            else:  # pragma: no cover
                pass

            previous_count = current_count

    def write(self):
        """After resolving all constraints. Write out all .unlock files

        :returns: Generator of ``.unlock`` absolute paths
        :rtype: collections.abc.Generator[pathlib.Path, None, None]
        """
        if is_module_debug:  # pragma: no cover
            msg_info = f"InFiles.write zeroes count: {len(list(self.zeroes))}"
            _logger.info(msg_info)
        else:  # pragma: no cover
            pass

        for in_ in self.zeroes:
            if in_.stem != "pins":
                abspath_zero = in_.abspath(self.cwd)
                file_name = f"{in_.stem}{SUFFIX_UNLOCKED}"
                abspath_unlocked = abspath_zero.parent.joinpath(file_name)

                if is_module_debug:  # pragma: no cover
                    msg_info = f"InFiles.write abspath_unlocked: {abspath_unlocked}"
                    _logger.info(msg_info)
                else:  # pragma: no cover
                    pass

                abspath_unlocked.touch(mode=0o644, exist_ok=True)
                is_file = abspath_unlocked.exists() and abspath_unlocked.is_file()

                if is_module_debug:  # pragma: no cover
                    _logger.info(f"InFiles.write is_file: {is_file}")
                else:  # pragma: no cover
                    pass

                if is_file:
                    contents = "\n".join(list(in_.requirements))
                    contents = f"{contents}\n"
                    abspath_unlocked.write_text(contents)
                    yield abspath_unlocked
                else:  # pragma: no cover
                    pass
            else:  # pragma: no cover
                pass

        yield from ()


def unlock_compile(inst):
    """pip requirement files can contain both ``-r`` and ``-c`` lines.
    Relative path to requirement files and constraint files respectively.

    Originally thought ``-c`` was a :command:`pip-compile` convention,
    not a pip convention. Opps!

    With presence of both ``.in`` and ``.lock`` files and then using the
    ``.in`` files would imply the package is (dependency) unlocked.

    So ``.in`` files are ``.unlock`` files.

    Creating ``.unlock`` files would serve no additional purpose, besides
    being explicit about the state of the package. That the author probably
    is no longer actively maintaining the package and thus has purposefully
    left the dependencies unlocked.

    :param inst:

       Backend subclass instance which has folders property containing
       ``collections.abc.Sequence[pathlib.Path]``

    :type inst: BackendType
    :returns: Generator of abs path to .lock files
    :rtype: collections.abc.Generator[pathlib.Path, None, None]
    :raises:

       - :py:exc:`drain_swamp.exceptions.MissingRequirementsFoldersFiles` --
         there are unresolvable constraint(s)

    """
    # Look at the folders. Then convert all ``.in`` --> ``.unlock``
    path_cwd = inst.parent_dir
    gen_unlocked_files = inst.in_files()
    in_files = list(gen_unlocked_files)
    del gen_unlocked_files

    # read in all .in files. key path_abs
    try:
        files = InFiles(path_cwd, in_files)
        files.resolution_loop()
    except MissingRequirementsFoldersFiles:
        raise
    else:
        gen = files.write()
        lst_called = list(gen)
        for abspath in lst_called:
            assert abspath.exists() and abspath.is_file()

        yield from lst_called

    yield from ()


def refresh_links(inst, is_set_lock=None):
    """Create/refresh ``.lnk`` files

    Does not write .lock or .unlock files

    :param inst:

       Backend subclass instance which has folders property containing
       ``collections.abc.Sequence[pathlib.Path]``

    :type inst: BackendType
    :param is_set_lock:

       Force the dependency lock. True to lock. False to unlock. None
       to use current lock state

    :type is_set_lock: bool | None
    :raises:

       - :py:exc:`drain_swamp.exceptions.MissingRequirementsFoldersFiles` --
         there are unresolvable constraint(s)

       - :py:exc:`AssertionError` -- In pyproject.toml no section,
         tool.setuptools.dynamic

       - :py:exc:`drain_swamp.exceptions.PyProjectTOMLParseError` --
         either not found or cannot be parsed

       - :py:exc:`drain_swamp.exceptions.PyProjectTOMLReadError` --
         Either not a file or lacks read permission

       - :py:exc:`TypeError` -- is_set_lock unsupported type expecting None or bool

    """
    is_invalid_set_lock = is_set_lock is not None and not isinstance(is_set_lock, bool)
    if is_invalid_set_lock:
        msg_exc = (
            "refresh_links parameter is_set_lock can be either None or "
            f"a boolean, got {is_set_lock!r}"
        )
        raise TypeError(msg_exc)
    else:  # pragma: no cover
        pass

    mod_path = "lock_toggle.refresh_links"
    path_cwd = inst.parent_dir
    path_config = inst.path_config
    gen_unlocked_files = inst.in_files()
    in_files = list(gen_unlocked_files)
    del gen_unlocked_files

    if is_module_debug:  # pragma: no cover
        msg_info = f"{mod_path} in_files {in_files}"
        _logger.info(msg_info)
    else:  # pragma: no cover
        pass

    # read in all .in files. key path_abs
    try:
        files = InFiles(path_cwd, in_files)
        files.resolution_loop()
    except MissingRequirementsFoldersFiles:
        raise

    if is_set_lock is None:
        # Get dependency lock state from pyproject.toml
        try:
            is_locked = inst.is_locked(path_config)
        except Exception:
            # PyProjectTOMLParseError, PyProjectTOMLReadError, AssertionError
            raise
    else:
        """To update symlinks to unlock dependencies

        .. code-block:: shell

           python -m build --config-setting="--set-lock=1" --sdist

        To update symlinks to lock dependencies

        .. code-block:: shell

           python -m build --config-setting="--set-lock=0" --sdist

        """
        is_locked = is_set_lock
    suffix = SUFFIX_LOCKED if is_locked else SUFFIX_UNLOCKED

    if is_module_debug:  # pragma: no cover
        msg_info = (
            f"{mod_path} is_set_lock --> is_locked {is_set_lock!r} --> {is_locked!r}"
        )
        _logger.info(msg_info)
        zeroes_count = len(list(files.zeroes))
        msg_info = f"{mod_path} files.zeroes zeroes_count {zeroes_count}"
        _logger.info(msg_info)
        msg_info = f"{mod_path} files: {files}"
        _logger.info(msg_info)
    else:  # pragma: no cover
        pass

    msg_warn = (
        "{}. No corresponding .unlock / .lock files"
        "Cannot make symlink. "
        "In {}, prepare the missing folders and files"
    )
    for in_ in files.zeroes:
        if in_.stem == "pins":
            # pins.in is used as-is
            continue
        else:  # pragma: no cover
            if is_module_debug:  # pragma: no cover
                msg_info = f"{mod_path} files.zeroes InFile {in_}"
                _logger.info(msg_info)
            else:  # pragma: no cover
                pass
            abspath_zero = in_.abspath(path_cwd)
            file_name = f"{in_.stem}{suffix}"
            abspath = abspath_zero.parent.joinpath(file_name)
            is_dest_file_exists = abspath.exists() and abspath.is_file()
            if not is_dest_file_exists:
                # No dest file, so skip creating a symlink
                msg_warn_missing_dest = msg_warn.format(abspath, path_cwd)
                raise MissingRequirementsFoldersFiles(msg_warn_missing_dest)
            else:
                if is_module_debug:  # pragma: no cover
                    msg_info = f"{mod_path} path_cwd: {path_cwd}"
                    _logger.info(msg_info)
                    msg_info = f"{mod_path} abspath: {abspath}"
                    _logger.info(msg_info)
                else:  # pragma: no cover
                    pass

                try:
                    _maintain_symlink(path_cwd, abspath)
                except (NotADirectoryError, FileNotFoundError) as e:
                    msg_warn_missing_dest = msg_warn.format(abspath, path_cwd)
                    raise MissingRequirementsFoldersFiles(msg_warn_missing_dest) from e
                except Exception as e:  # pragma: no cover
                    # Sad, but not the end of the world
                    msg_exc = str(e)
                    _logger.warning(msg_exc)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Mundane :py:func:`subprocess.run` call. Blocking and not multiprocessing

.. py:data:: __all__
   :type: tuple[str]
   :value: ("run_cmd",)

   Module exports

"""

import os
import subprocess
from collections.abc import Sequence
from pathlib import (
    Path,
    PurePath,
)

__all__ = ("run_cmd",)


def run_cmd(cmd, cwd=None, env=None):
    """Run cmd in subprocess, capture both stdout and stderr

    :param cmd: command to run in a subprocess
    :type cmd: collections.abc.Sequence[str]
    :param cwd: Default None
    :type cwd: pathlib.Path | None
    :param env: Default None. Expecting an :py:class:`os._Environ`
    :type env: typing.Any | None
    :returns: log messages, exception messages, return code, subprocess failure message
    :rtype: tuple[str | None, str | None, int | None, str | None]
    :raises:

        - :py:exc:`TypeError` -- Unsupported type for 1st arg cmd

    """
    if not isinstance(cmd, Sequence):
        msg_warn = f"expected 1st param cmd to be a Sequence got {type(cmd)}"
        raise TypeError(msg_warn)

    if cwd is None or not issubclass(type(cwd), PurePath):
        path_cwd = Path.cwd()
    else:  # pragma: no cover
        path_cwd = cwd

    is_not_env = env is None or not isinstance(env, os._Environ)
    if is_not_env:
        opt_env = None
    else:  # pragma: no cover
        opt_env = env

    try:
        proc = subprocess.run(
            cmd,
            shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=path_cwd,
            env=opt_env,
            text=True,
        )
    except OSError as e:
        # e.g. cmd=("bin/true",)
        str_err = str(f"{e.strerror} {e.filename}")
        ret = (None, None, None, str_err)
    except (
        subprocess.CalledProcessError,
        subprocess.TimeoutExpired,
    ) as e:  # pragma: no cover
        """setuptools-scm requires at least one commit. Could not get
        semantic version"""
        # DO NOT USE proc.stdout, differing meaning
        str_err = str(e.stderr)
        str_out = None
        exit_code = None
        ret = (str_out, None, exit_code, str_err)
    else:
        # warning level log messages bleed to stdout
        str_out = proc.stdout
        if str_out is None or len(str_out.rstrip()) == 0:
            str_out = None
        else:
            str_out = str_out.rstrip()

        # Exception details written to stderr
        str_err = proc.stderr
        if str_err is None or len(str_err.rstrip()) == 0:
            str_err = None
        else:
            str_err = str_err.rstrip()

        # str_out is only Warning level log messages
        ret = (str_out, str_err, proc.returncode, None)

    return ret



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

Module to construct setuptools-scm PRETEND_VERSION environment variable string

.. py:data:: PRETEND_KEY
   :type: str
   :value: "SETUPTOOLS_SCM_PRETEND_VERSION"

.. py:data:: PRETEND_KEY_NAMED
   :type: str
   :value: "SETUPTOOLS_SCM_PRETEND_VERSION_for_{name}"

   Get environmental variable key for forcing setuptools-scm to use a
   particular semantic version

.. py:data:: __all__
   :type: tuple[str]
   :value: ("normalize_dist_name")

   Module exports

.. seealso::

   `[Source] <https://github.com/pypa/setuptools_scm/blob/main/src/setuptools_scm/_overrides.py>`_
   `[LICENSE:MIT] <https://github.com/pypa/setuptools_scm/blob/main/LICENSE>`_

"""

import os
import re

PRETEND_KEY = "SETUPTOOLS_SCM_PRETEND_VERSION"
PRETEND_KEY_NAMED = PRETEND_KEY + "_FOR_{name}"

__all__ = ("normalize_dist_name",)


def normalize_dist_name(dist_name):
    """Normalize the dist name as per PEP 503

    :param dist_name:

       May contain period(s), hyphen(s), and/or underscore(s)

    :type dist_name: str
    :returns: dist name normalized and uppercase
    :rtype: str
    """
    normalized_dist_name = re.sub(r"[-_.]+", "-", dist_name)
    env_var_dist_name = normalized_dist_name.replace("-", "_").upper()

    return env_var_dist_name


def _scm_key(dist_name):
    """When want to set a specific version, setuptools-scm offers an
    environment variable which overrides normal behavior

    This is needed when wanting to create a:
    - tagged version
    - post-release version
    - pre-release version

    :param dist_name:

       package name. Will upper case and replace hyphens with underscores

    :type dist_name: str
    :returns: environment variable to affect setuptools-scm behavior
    :rtype: str
    :meta private:
    """
    # source setuptools_scm._override.read_named_env
    env_var_dist_name = normalize_dist_name(dist_name)
    d_named = {"name": env_var_dist_name}
    scm_override_key = PRETEND_KEY_NAMED.format(**d_named)

    return scm_override_key


def read_named_env(
    *,
    tool="SETUPTOOLS_SCM",
    name,
    dist_name,
):
    """From environment, get variable value. Will contain the version

    :param tool: Default "SETUPTOOLS_SCM". Hardcoded tool name
    :type tool: str
    :param name: Can be ``PRETEND_VERSION`` or ``OVERRIDES``
    :type name: str
    :param dist_name: target package name
    :type dist_name: str | None
    :returns: version str
    :rtype: str | None
    """
    if dist_name is not None:
        # Normalize the dist name as per PEP 503.
        env_var_dist_name = normalize_dist_name(dist_name)
        val = os.environ.get(f"{tool}_{name}_FOR_{env_var_dist_name}")
        if val is not None:
            ret = val
        else:  # pragma: no cover
            ret = None
    else:  # pragma: no cover
        ret = None

    if ret is None:
        ret = os.environ.get(f"{tool}_{name}")
    else:  # pragma: no cover
        pass

    return ret



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_snippet_sphinx_conf.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_snippet_sphinx_conf.py

Needs a config file to specify exact files to include / omit from report.
Will fail with exit code 1 even with 100% coverage

"""

import subprocess
from datetime import datetime
from pathlib import (
    Path,
    PurePath,
)
from unittest.mock import patch

import pytest

from drain_swamp.constants import g_app_name
from drain_swamp.snippet_sphinx_conf import SnipSphinxConf

testdata_snip_sphinx_conf_init = (
    pytest.param(None, marks=pytest.mark.xfail(raises=NotADirectoryError)),
    pytest.param("doc", marks=pytest.mark.xfail(raises=FileNotFoundError)),
    pytest.param("docs", marks=pytest.mark.xfail(raises=FileNotFoundError)),
)
ids_snip_sphinx_conf_init = (
    "No doc/ or docs/ sub-folder",
    "doc/ sub-folder",
    "docs/ sub-folder",
)


@pytest.mark.parametrize(
    "a_folder",
    testdata_snip_sphinx_conf_init,
    ids=ids_snip_sphinx_conf_init,
)
def test_snip_sphinx_conf_init(a_folder, tmp_path):
    """Sphinx [sub-folder]/conf.py does not exist exceptions"""
    if a_folder is not None and isinstance(a_folder, str):
        path_docs = tmp_path.joinpath(a_folder)
        path_docs.mkdir()

    SnipSphinxConf(path=tmp_path)


testdata_now_to_str = (
    pytest.param(None, marks=pytest.mark.xfail(raises=TypeError)),
    pytest.param(1.2345, marks=pytest.mark.xfail(raises=TypeError)),
)
ids_now_to_str = (
    "None",
    "float",
)


@pytest.mark.parametrize(
    "format_str",
    testdata_now_to_str,
    ids=ids_now_to_str,
)
def test_now_to_str(format_str, path_project_base):
    assert isinstance(SnipSphinxConf.now(), datetime)
    # strftime / strptime format str
    SnipSphinxConf.now_to_str(format_str)


def test_snip_sphinx_conf_properties(path_project_base):
    sc = SnipSphinxConf(path=path_project_base)

    # SnipSphinxConf.now getter
    assert isinstance(SnipSphinxConf.now(), datetime)

    assert issubclass(type(sc.path_abs), PurePath)

    # SnipSphinxConf.path_abs getter
    #     contains a ``conf.py`` file
    assert sc.path_abs.exists() and sc.path_abs.is_file()

    # SnipSphinxConf.path_cwd getter
    doc_folders = list(sc.path_cwd.glob("doc?"))
    doc_folder_count = len(doc_folders)
    assert doc_folder_count == 1

    assert sc.SV is None
    assert sc.author_name_left is None


test_data_contents = (
    ("0.0.1", g_app_name, "1850", "0.0.1"),
    ("now", g_app_name, "1850", "0.0.1"),
    ("current", g_app_name, "1850", "0.0.1"),
    ("tag", g_app_name, "1850", "0.0.1"),
)
ids_data_contents = (
    "kind -- explicit version str",
    "kind -- current alias",
    "kind -- current",
    "kind -- tag, fallback current",
)


@pytest.mark.parametrize(
    "kind, package_name, copyright_start_year, sem_version_str",
    test_data_contents,
    ids=ids_data_contents,
)
def test_snip_sphinx_conf_contents(
    kind,
    package_name,
    copyright_start_year,
    sem_version_str,
    path_project_base,
):
    cmd = []
    sc = SnipSphinxConf(path=path_project_base)

    # properties; defang so doesn't change src/[proj name]/_version.py
    with (
        patch(
            "subprocess.run",
            return_value=subprocess.CompletedProcess(
                cmd,
                returncode=0,
                stdout=sem_version_str,
            ),
        ),
    ):
        sc.contents(
            kind,
            package_name,
            copyright_start_year,
        )
        actual = sc._contents

        assert actual is not None
        assert isinstance(actual, str)

        # From package metadata
        assert "Dave Faulkmore" in actual

        # Has and sets these Sphinx variables
        assert """copyright = \"""" in actual
        #    xyz version
        assert """version = \"""" in actual
        #    semantic version
        assert """release = \"""" in actual
        #    Terse datetime str intended for display
        assert """release_date = \"""" in actual

    # Could not get project name from git --> AssertionError
    """
    if kind == "tag":
        with (
            patch(
                f"{g_app_name}.version_semantic._get_app_name",
                side_effect=AssertionError,
            ),
            patch(
                "subprocess.run",
                return_value=subprocess.CompletedProcess(
                    cmd,
                    returncode=128,
                    stdout="",
                    stderr="fatal: No names found, cannot describe anything.",
                ),
            ),
            pytest.raises(AssertionError),
        ):
            sc.contents(
                kind,
                package_name,
                copyright_start_year,
            )
    """
    pass

    # Explicit version str bad --> ValueError
    kind = "nowhere boys alternative universe police state that knows about magic"
    with pytest.raises(ValueError):
        sc.contents(
            kind,
            package_name,
            copyright_start_year,
        )


testdata_snip_sphinx_conf_replace = (
    (
        Path(__file__).parent.joinpath(
            "test_snip",
            "test_snip_harden_one_snip__with_id_.txt",
        ),
        "0.0.1",
        g_app_name,
        "1850",
        "asdf",
        """zzzzzzzzzzzzz
# @@@ editable {}
{}
# @@@ end
zzzzzzzzzzzzz""",
    ),
)
ids_snip_sphinx_conf_replace = ("replace compare against template",)


@pytest.mark.parametrize(
    "path_example, kind, package_name, copyright_start_year, id_, skeleton",
    testdata_snip_sphinx_conf_replace,
    ids=ids_snip_sphinx_conf_replace,
)
def test_snip_sphinx_conf_replace(
    path_example,
    kind,
    package_name,
    copyright_start_year,
    id_,
    skeleton,
    tmp_path,
):
    path_dir = tmp_path.joinpath("docs")
    path_dir.mkdir()

    contents_orig = path_example.read_text()

    path_file = path_dir.joinpath("conf.py")
    path_file.write_text(contents_orig)

    sc = SnipSphinxConf(path=tmp_path)
    assert sc.path_abs == path_file

    cmd = []

    with (
        patch(
            "subprocess.run",
            return_value=subprocess.CompletedProcess(
                cmd,
                returncode=0,
                stdout=kind,
            ),
        ),
    ):
        sc.contents(
            kind,
            package_name,
            copyright_start_year,
        )
        contents = sc._contents

    expected = skeleton.format(id_, contents.rstrip())

    sc.replace(snippet_co=id_)
    actual = sc.path_abs.read_text()

    assert expected.rstrip() == actual.rstrip()



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_cli_scm_version.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_cli_scm_version.py

"""

import logging
import logging.config
from pathlib import Path
from unittest.mock import patch

import pytest  # noqa: F401
from click.testing import CliRunner

from drain_swamp.cli_scm_version import (
    entrypoint_name,
    get_scm_version,
    main,
    write_scm_version,
)
from drain_swamp.constants import (
    LOGGING,
    g_app_name,
)


def test_cli_main():
    """Minimally test package version is printed"""
    runner = CliRunner()
    # --version
    """
    cmd = ["--version"]
    result = runner.invoke(main, cmd)
    assert result.exit_code == 0
    assert "version" in result.stdout
    """

    # --help
    cmd = ["--help"]
    result = runner.invoke(main, cmd)
    assert result.exit_code == 0
    assert f"Command-line for {entrypoint_name}. Prints usage" in result.stdout


def test_get_scm_version_missing_pyproject_toml(tmp_path, prepare_folders_files):
    # pytest --showlocals --log-level INFO -k "test_get_scm_version_missing_pyproject_toml" tests
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        seq_rel_path = ("_version.py",)
        path_version_file = path_tmp_dir.joinpath(seq_rel_path[0])

        # prepare
        prepare_folders_files(seq_rel_path, path_tmp_dir)

        # no pyproject.toml --> LookupError (3). Missing sections
        scm_ver = "0.0.1"
        cmd = (scm_ver, "--path", path_tmp_dir, "--write-to", seq_rel_path[0])
        result = runner.invoke(write_scm_version, cmd)
        assert result.exit_code == 3
        assert path_version_file.stat().st_size == 0
        del scm_ver

        # no pyproject.toml --> LookupError (3). Missing sections
        cmd = ("--path", path_tmp_dir, "--is-write", "--write-to", seq_rel_path[0])
        result = runner.invoke(get_scm_version, cmd)
        assert result.exit_code == 3


def test_get_scm_version_normal(
    tmp_path, prep_pyproject_toml, prepare_folders_files, caplog, has_logging_occurred
):
    """How to simulate reading version from git?

    .. todo:: is git

       skip if no git
       skip if git not initialized

    """
    # pytest --showlocals --log-level INFO -k "test_get_scm_version_normal" tests
    LOGGING["loggers"][g_app_name]["propagate"] = True
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(name=g_app_name)
    logger.addHandler(hdlr=caplog.handler)
    caplog.handler.level = logger.level

    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path) as tmp_dir_path:
        path_tmp_dir = Path(tmp_dir_path)
        version_file = "src/complete_awesome_perfect/_version.py"
        seq_rel_path = (version_file,)
        path_version_file = path_tmp_dir.joinpath(seq_rel_path[0])

        # prepare
        #    pyproject.toml
        p_toml_file = Path(__file__).parent.joinpath(
            "_good_files", "complete-manage-pip-prod-unlock.pyproject_toml"
        )
        prep_pyproject_toml(p_toml_file, path_tmp_dir)
        #    version_file folders and empty file
        prepare_folders_files(seq_rel_path, path_tmp_dir)

        #    version file (skeleton/minimal)
        str_version_file = (
            """__version__ = version = '0.2.0'\n"""
            """__version_tuple__ = version_tuple = (0, 2, 0)\n\n"""
        )
        path_version_file.write_text(str_version_file)

        # act -- get_scm_version
        cmd = ("--path", path_tmp_dir, "--is-write", "--write-to", seq_rel_path[0])
        result = runner.invoke(get_scm_version, cmd)
        assert result.exit_code == 0
        #    repo scm version
        assert len(result.output) != 0
        scm_ver = result.output
        #    wrote to version_file
        assert path_version_file.stat().st_size != 0
        path_version_file.unlink()

        # act -- write_scm_version -- valid semantic version str
        cmd = (scm_ver, "--path", path_tmp_dir, "--write-to", seq_rel_path[0])
        result = runner.invoke(write_scm_version, cmd)
        assert result.exit_code == 0
        assert path_version_file.stat().st_size != 0
        path_version_file.unlink()

        # version_file from pyproject.toml
        cmd = (scm_ver, "--path", path_tmp_dir)
        result = runner.invoke(write_scm_version, cmd)
        assert result.exit_code == 0
        assert path_version_file.stat().st_size != 0
        path_version_file.unlink()

        # invalid semantic version str -->  ValueError(4)
        scm_ver = "golf balls get lost"
        cmd = (scm_ver, "--path", path_tmp_dir)
        result = runner.invoke(write_scm_version, cmd)
        assert result.exit_code == 4

        """upstream bug. Passes regex, but invalid semantic version str.
        exception not caught in setuptools_scm.version.tag_to_version

        Exception properly handled in setuptools_scm._version_cls._version_as_tuple"""
        with patch(
            "drain_swamp.cli_scm_version.scm_version",
            return_value="0.0.1-dev1.g1234123",
        ):
            cmd = ("--path", path_tmp_dir, "--is-write", "--write-to", seq_rel_path[0])
            result = runner.invoke(get_scm_version, cmd)
            assert result.exit_code == 4

    assert has_logging_occurred(caplog)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_version_file_overrides.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_version_file_overrides.py

"""

import os
from unittest.mock import patch

import pytest

from drain_swamp.version_file._overrides import (
    _scm_key,
    normalize_dist_name,
    read_named_env,
)

testdata_normalize_dist_name = (
    (
        "awesome.special_project",
        "AWESOME_SPECIAL_PROJECT",
    ),
    (
        "awesome.special-project-",
        "AWESOME_SPECIAL_PROJECT_",
    ),
)
ids_normalize_dist_name = (
    "period and underline",
    "period and hyphens. Trailing hyphen",
)


@pytest.mark.parametrize(
    "dist_name, expected",
    testdata_normalize_dist_name,
    ids=ids_normalize_dist_name,
)
def test_normalize_dist_name(dist_name, expected):
    actual = normalize_dist_name(dist_name)
    assert actual == expected


def test_read_named_env():
    env_0 = os.environ.copy()
    dist_name = "AWESOME_SPECIAL_PROJECT"
    scm_ver = "0.0.1"

    # PRETEND_VERSION -- env variable exists
    key = _scm_key(dist_name)
    env_0[key] = scm_ver
    with patch("os.environ.get", wraps=env_0.get):
        actual_val = read_named_env(name="PRETEND_VERSION", dist_name=dist_name)
    assert actual_val == scm_ver

    # OVERRIDES -- no dist_name
    dist_name = None
    env_1 = os.environ.copy()
    key = "SETUPTOOLS_SCM_PRETEND_VERSION"
    env_1[key] = scm_ver
    with patch("os.environ.get", wraps=env_1.get):
        actual_val = read_named_env(name="PRETEND_VERSION", dist_name=None)
    assert actual_val == scm_ver



from pathlib import Path
from typing import Final

entrypoint_name: Final[str]
help_path: Final[str]
help_is_write: Final[str]
help_write_to: Final[str]
EPILOG_SCM_VERSION_GET: Final[str]
EPILOG_SCM_VERSION_WRITE: Final[str]

def main() -> None: ...
def get_scm_version(path: Path, is_write: bool, write_to: str | None) -> None: ...
def write_scm_version(scm_ver: str, path: Path, write_to: str | None) -> None: ...



from collections.abc import Sequence
from pathlib import Path
from typing import Any

__all__ = (
    "is_ok",
    "is_relative_required",
    "is_iterable_not_str",
)

def is_ok(test: Any | None) -> bool: ...
def is_iterable_not_str(mixed: Any | None) -> bool: ...
def is_relative_required(
    path_relative: Path | None = None,
    extensions: Sequence[str] = ...,
) -> bool: ...
def click_bool(val: str | None = None) -> bool | None: ...



from pathlib import Path

__all__ = (
    "dump_version",
    "write_version_to_path",
    "write_version_files",
)

TEMPLATES: dict[str, str]

def dump_version(
    root: Path,
    version: str,
    write_to: Path,
    template: str | None = None,
) -> None: ...
def _validate_template(
    target: Path,
    template: str | None,
) -> str: ...
def write_version_to_path(
    target: Path,
    template: str | None,
    version: str,
) -> None: ...
def write_version_files(
    version: str,
    root: Path,
    write_to: str | None,
    version_file: str | None,
    is_only_not_exists: bool | None = False,
) -> None: ...



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

The _version module is generated and not guaranteed to exist.

When _version is missing, such as after a :code:`git clone` or within CI/CD,
there is no reason to allow this fyi only package info to be the cause of failure

.. py:data:: __version_app
   :type: str

   Official app version. Semantic versioning fully supported. So pre and post releases,
   release candidate, and dev releases are possible

   tagged post releases are encouraged. Cuz that means used
   :command:`git rebase` to fix a problem at when it originally occured
   rather than applying the fix and ignoring the past

.. py:data:: __url__
   :type: str

   rtd URL to for current tagged version. This is not used for anything.
   Just here for completeness.

.. py:data:: __all__
   :type: tuple[str, str]
   :value: ("__version_app", "__url__")

   Module exports

.. todo:: lookup module

   Lookup in pyproject.toml for the dotted path of version file. Then
   atttempt to import it

"""

from .constants import g_app_name
from .version_semantic import (
    SemVersion,
    sanitize_tag,
)

__all__ = (
    "__version_app",
    "__url__",
)

try:
    # hardcoded import is bad
    from ._version import __version__
except (ModuleNotFoundError, ImportError):  # pragma: no cover
    __version_app = "unreleased"
    __url__ = None
else:
    # Removes epoch and local. Fixes version
    __version_app, local = sanitize_tag(__version__)

    sv = SemVersion()
    sv.parse_ver(__version__, local=local)
    __url__ = sv.readthedocs_url(g_app_name, is_latest=False)



from collections.abc import Mapping
from pathlib import Path
from typing import Any

__all__ = ("TomlParser",)

class TomlParser:
    def __init__(
        self,
        path: Any | None,
        raise_exceptions: Any | None = False,
    ) -> None: ...
    @property
    def path_file(self) -> Path | None: ...
    @property
    def d_pyproject_toml(self) -> Mapping[str, Any] | None: ...
    @classmethod
    def resolve(cls, path_config: Any | None) -> Path: ...
    def _get_pyproject_toml(self, path_config: Any | None) -> Mapping[str, Any]: ...



from .exceptions import (
    BackendNotSupportedError,
    MissingRequirementsFoldersFiles,
    PyProjectTOMLParseError,
    PyProjectTOMLReadError,
)

__all__ = (
    "PyProjectTOMLParseError",
    "BackendNotSupportedError",
    "PyProjectTOMLReadError",
    "MissingRequirementsFoldersFiles",
)



import subprocess
import unittest
from importlib.metadata import PackageNotFoundError
from pathlib import Path
from unittest.mock import patch

from drain_swamp.constants import g_app_name
from drain_swamp.version_semantic import (
    SemVersion,
    Version,
    _current_tag,
    _current_version,
    _get_app_name,
    _is_setuptools_scm,
    _map_release,
    _remove_v,
    get_version,
    sanitize_tag,
)

testdata_v = (
    ("v1.0.1", "1.0.1"),
    ("0!v1.0.1", "0!1.0.1"),
    ("1!v1.0.1", "1!1.0.1"),
    ("0!v1.0.1+g4b33a80.d20240129", "0!1.0.1+g4b33a80.d20240129"),
    ("1!v1.0.1+g4b33a80.d20240129", "1!1.0.1+g4b33a80.d20240129"),
    ("v0.1.1.dev0+g4b33a80.d20240129", "0.1.1.dev0+g4b33a80.d20240129"),
    ("v0.1.1.post0+g4b33a80.d20240129", "0.1.1.post0+g4b33a80.d20240129"),
    ("v0.1.1.a1dev1+g4b33a80.d20240129", "0.1.1.a1dev1+g4b33a80.d20240129"),
    ("v0.1.1.alpha1dev1+g4b33a80.d20240129", "0.1.1.alpha1dev1+g4b33a80.d20240129"),
    ("v0.1.1.b1dev1+g4b33a80.d20240129", "0.1.1.b1dev1+g4b33a80.d20240129"),
    ("v0.1.1.beta1dev1+g4b33a80.d20240129", "0.1.1.beta1dev1+g4b33a80.d20240129"),
    ("v0.1.1.rc1dev1+g4b33a80.d20240129", "0.1.1.rc1dev1+g4b33a80.d20240129"),
)

testdata_valids = (
    ("0.0.1", "0.0.1", "0-0-1"),  # tagged final version
    ("0.1.1.dev0+g4b33a80.d20240129", "0.1.1.dev0", "0-1-1"),
    ("0.1.1.dev1+g4b33a80.d20240129", "0.1.1.dev1", "0-1-1"),
    ("0.1.1.post0+g4b33a80.d20240129", "0.1.1.post0", "0-1-1post0"),
    ("0.1.1.a1dev1+g4b33a80.d20240129", "0.1.1a1.dev1", "0-1-1a1"),
    ("0.1.1.alpha1dev1+g4b33a80.d20240129", "0.1.1a1.dev1", "0-1-1a1"),
    ("0.1.1.b1dev1+g4b33a80.d20240129", "0.1.1b1.dev1", "0-1-1b1"),
    ("0.1.1.beta1dev1+g4b33a80.d20240129", "0.1.1b1.dev1", "0-1-1b1"),
    ("0.1.1.rc1dev1+g4b33a80.d20240129", "0.1.1rc1.dev1", "0-1-1rc1"),
)

testdata_invalids = (("0.1.1.candidate1dev1+g4b33a80.d20240129", "0.1.1rc1.dev1"),)

# long|short form --> long form
testdata_releaselevel = (
    ("candidate", "candidate"),
    ("rc", "candidate"),
    ("post", "post"),
    ("a", "alpha"),
    ("alpha", "alpha"),
    ("b", "beta"),
    ("beta", "beta"),
)


class PackageInstalled(unittest.TestCase):
    def test_is_setuptools_scm(self):
        """Checks whether or not package setuptools-scm is installed"""
        # non-existent nor installed package
        with patch(
            "importlib.metadata.version",
            side_effect=PackageNotFoundError,
        ):
            bol_actual = _is_setuptools_scm()
            self.assertFalse(bol_actual)


class PackageVersioning(unittest.TestCase):
    def setUp(self):
        self.vals = (
            (
                (0, 0, 1),
                {"releaselevel": "alpha", "serial": 0, "dev": None},
                "0.0.1a0",
            ),
            (
                (0, 0, 1),
                {"releaselevel": "beta", "serial": 0, "dev": None},
                "0.0.1b0",
            ),
            (
                (0, 0, 1),
                {"releaselevel": "candidate", "serial": 0, "dev": None},
                "0.0.1rc0",
            ),
            (
                (0, 0, 1),
                {"releaselevel": "", "serial": 0, "dev": None},
                "0.0.1",
            ),
            (
                (0, 0, 1),
                {"releaselevel": "alpha", "serial": 3, "dev": 10},
                "0.0.1a3.dev10",
            ),
            (
                (0, 0, 1),
                {"releaselevel": "post", "serial": 3, "dev": None},
                "0.0.1post3",
            ),
        )

    def test_setuptools_scm_version_file(self):
        """Autogenerated file by setuptools-scm"""
        valids = (
            ((0, 5, 2), "0.5.2"),
            ((0, 5, 2, "dev0"), "0.5.2.dev0"),
            ((0, 5, 2, "dev0", "g1234567.d20240101"), "0.5.2.dev0+g1234567.d20240101"),
        )
        for t_ver, ver_expected in valids:
            ver = list(t_ver[:3])
            ver_short = ".".join(map(str, ver))
            if len(t_ver) == 3:
                # 0.0.1
                ver_long = ver_short
                ver_actual = ver_long
            elif len(t_ver) in [4, 5]:
                # __version_tuple__
                # 0.0.1.a1dev8
                ver_dev = t_ver[3]
                ver_long = ver_short
                if len(ver_dev) != 0:
                    ver_long += f".{ver_dev}"

                if len(t_ver) == 4:
                    ver_actual = ver_long
                else:
                    ver_git = t_ver[-1]
                    ver_actual = f"{ver_long}+{ver_git}"

            self.assertEqual(ver_actual, ver_expected)

    def test_sanitize_tag(self):
        """Convert repo version --> semantic version"""
        for v_in, v_expected, anchor in testdata_valids:
            v_actual, _ = sanitize_tag(v_in)
            self.assertEqual(v_actual, v_expected)

        # sanitize_tag and get_version both should fail
        for v_in, v_expected in testdata_invalids:
            with self.assertRaises(ValueError):
                sanitize_tag(v_in)

        for v_in, v_expected in testdata_invalids:
            with self.assertRaises(ValueError):
                get_version(v_in, is_use_final=True)

        # Strip epoch and locals
        valids = (
            ("1!1.0.1a1.dev1", "1.0.1a1.dev1"),
            ("1.0.1a1.dev1+4b33a80.4b33a80", "1.0.1a1.dev1"),
        )
        for orig, expected in valids:
            actual, _ = sanitize_tag(orig)
            self.assertEqual(actual, expected)

    def test_get_version(self):
        """Used for display only. Allows release level, final"""
        # Flip the logic backwards
        finals = (
            None,
            0.12345,  # unsupported type
            False,
        )
        for final in finals:
            for args, kwargs, actual in self.vals:
                expect_info, expect_dev = get_version(
                    actual,
                    is_use_final=final,
                )
                self.assertEqual(kwargs["dev"], expect_dev)
                self.assertEqual(kwargs["serial"], expect_info[-1])
                if len(kwargs["releaselevel"]) == 0:
                    self.assertEqual(len(expect_info[-2]), 0)
                else:
                    self.assertEqual(kwargs["releaselevel"], expect_info[-2])
                self.assertEqual(args, expect_info[:3])

        # Allow final
        actual = "1.0.1"
        l_actual = actual.split(".")
        l_actual2 = map(int, iter(l_actual))
        t_actual = tuple(l_actual2)
        expect_info, expect_dev = get_version(
            actual,
            is_use_final=True,
        )
        self.assertIsNone(expect_dev)
        self.assertEqual(0, expect_info[-1])
        self.assertEqual("final", expect_info[-2])
        self.assertEqual(t_actual, expect_info[:3])

        # Has both dev and is a prerelease
        dev_pres = (
            ("0.1.1.a1dev1+g4b33a80.d20240129", "0.1.1a1.dev1"),
            ("0.1.1.alpha1dev1+g4b33a80.d20240129", "0.1.1a1.dev1"),
            ("0.1.1.b1dev1+g4b33a80.d20240129", "0.1.1b1.dev1"),
            ("0.1.1.beta1dev1+g4b33a80.d20240129", "0.1.1b1.dev1"),
            ("0.1.1.rc1dev1+g4b33a80.d20240129", "0.1.1rc1.dev1"),
        )

        for dev_pre in dev_pres:
            expected, _ = sanitize_tag(dev_pre[1])
            expect_info, expect_dev = get_version(expected)

            v = Version(expected)

            v_pre = v.pre
            v_pre_is = v.is_prerelease
            v_dev = v.dev

            # long format
            pre = expect_info[-2]
            self.assertIn(pre, _map_release.keys())

            found_k = None
            for k, v in _map_release.items():
                if pre == k:
                    found_k = k

            self.assertEqual(expect_dev, v_dev)

            self.assertIsNotNone(found_k)
            # pre is long format. So ``alpha`` rather than ``a``
            self.assertEqual(pre, found_k)

        # Has dev and no releaselevel
        dev_pres = (
            ("0.1.1.a1dev1+g4b33a80.d20240129", "0.1.1.dev1"),
            ("0.1.1.alpha1dev1+g4b33a80.d20240129", "0.1.1.dev0"),
            ("0.1.1.b1dev1+g4b33a80.d20240129", "0.1.1dev8"),
        )
        for dev_pre in dev_pres:
            expected, _ = sanitize_tag(dev_pre[1])
            expect_info, expect_dev = get_version(expected)

            v = Version(expected)

            v_pre = v.pre
            v_pre_is = v.is_prerelease
            v_dev = v.dev
            self.assertEqual(expect_dev, v_dev)
            self.assertIsNone(v_pre)

        # post only
        dev_pres = (
            ("0.1.1.post0+g4b33a80.d20240129", "0.1.1.post0", 0),
            ("0.1.1.post8", "0.1.1.post8", 8),
            ("0.1.1.post5", "0.1.1post5", 5),
        )
        for dev_pre in dev_pres:
            expected_post = dev_pre[2]
            expected, _ = sanitize_tag(dev_pre[1])
            expect_info, expect_dev = get_version(expected)

            v = Version(expected)

            v_post = v.post
            v_post_is = v.is_postrelease
            v_pre_is = v.is_prerelease
            self.assertTrue(v_post_is)
            self.assertFalse(v_pre_is)
            self.assertEqual(v_post, expected_post)
            self.assertEqual(expect_info[-2], "post")

    def test_v_remove(self):
        """Remove epoch and local and v prefix from semantic version str"""
        for v_in, expected in testdata_v:
            actual = _remove_v(v_in)
            self.assertEqual(actual, expected)


class SemVersioning(unittest.TestCase):
    def setUp(self):
        if "__pycache__" in __file__:
            # cached
            self.path_tests = Path(__file__).parent.parent
        else:
            # not cached
            self.path_tests = Path(__file__).parent
        self.cwd = self.path_tests.parent

    def test_current_tag(self):
        """If not tagged version at all will be None"""
        cmd = []
        with patch(
            "subprocess.run",
            side_effect=subprocess.CalledProcessError(128, cmd),
        ):
            actual = _current_tag(path=self.cwd)
            self.assertIsNone(actual)

        # simulate git saying, no tagged versions yet
        with patch(
            "subprocess.run",
            return_value=subprocess.CompletedProcess(
                cmd,
                returncode=128,
                stderr="fatal: No names found, cannot describe anything.",
            ),
        ):
            actual = _current_tag(path=self.cwd)
            self.assertIsNone(actual)

        # Successful call
        expected = "0.0.1"
        with (
            patch(
                "subprocess.run",
                return_value=subprocess.CompletedProcess(
                    cmd,
                    returncode=0,
                    stdout=expected,
                ),
            ),
        ):
            actual = _current_tag(path=self.cwd)
            self.assertIsInstance(actual, str)
            self.assertEqual(actual, expected)

    def test_current_version(self):
        """setuptools-scm retrieves development version from git
        Without epoch, local, or prepended v
        """
        # setuptools-scm is not installed.
        #    Current version is obtained from package setuptools-scm
        with patch(
            f"{g_app_name}.version_semantic._is_setuptools_scm",
            return_value=False,
        ):
            actual = _current_version(path=self.cwd)
            self.assertIsNone(actual)

        cmd = []
        with patch(
            "subprocess.run",
            side_effect=subprocess.CalledProcessError(128, cmd),
        ):
            actual = _current_version(path=self.cwd)
            self.assertIsNone(actual)

        with patch(
            "subprocess.run",
            return_value=subprocess.CompletedProcess(
                cmd,
                returncode=128,
                stdout="",
            ),
        ):
            actual = _current_version(path=self.cwd)
            self.assertIsNone(actual)

        # Successful call
        expected = "0.0.1"
        with (
            patch(
                "subprocess.run",
                return_value=subprocess.CompletedProcess(
                    cmd,
                    returncode=0,
                    stdout=expected,
                ),
            ),
        ):
            actual = _current_version(path=self.cwd)
            self.assertIsInstance(actual, str)
            self.assertEqual(actual, expected)

    def test_semversion_properties(self):
        """Parse and check SemVersion properties"""
        sv = SemVersion()
        for v_in, expected in testdata_v:
            sv.parse_ver(v_in)
            self.assertIsInstance(sv.major, int)
            self.assertIsInstance(sv.minor, int)
            self.assertIsInstance(sv.micro, int)
            self.assertIsInstance(sv.releaselevel, str)
            self.assertIsInstance(sv.serial, int)
            if sv.dev is None:
                self.assertIsNone(sv.dev)
            else:
                self.assertIsInstance(sv.dev, int)

        # path_cwd setter
        path_dir = self.cwd
        path_pyproject_toml = path_dir / "pyproject.toml"
        with self.assertRaises(NotADirectoryError):
            SemVersion(path=path_pyproject_toml)

        # is_use_final setter
        invalids = (
            None,
            1.2345,
            "",
        )
        for invalid in invalids:
            sv.is_use_final = True
            sv.is_use_final = invalid
            self.assertFalse(sv.is_use_final)

    def test_semversion_parse_ver(self):
        """SemVersion now stores version local"""
        sv = SemVersion()
        locals_ = (
            (None, None),  # empty string or None
            ("", None),  # empty string or None
            ("    ", None),  # empty string or None
            ("+g2988c13e.d20240101", None),  # a local
            (1.12345, None),  # unsupported
        )
        for local, expected in locals_:
            for v_in, expected in testdata_v:
                sv.parse_ver(v_in, local=local)
                sv._local = expected

    def test_readthedocs_package_name(self):
        """rtd url just because"""
        project_name = g_app_name.replace("_", "-")
        sv = SemVersion()

        # parse_ver never called. ver None --> latest
        is_latests = (
            None,
            "latest",
            1.2345,
        )
        for is_latest in is_latests:
            str_url = sv.readthedocs_url(project_name, is_latest=is_latest)
            self.assertTrue(str_url.endswith("latest"))
            protocol_len = len("https://")
            uri = str_url[protocol_len:]
            project_name_actual = uri.split(".")[0]
            self.assertNotIn("_", project_name_actual)

        # package_name contains hyphens. Gets converted to underscores
        str_url = sv.readthedocs_url(g_app_name)
        protocol_len = len("https://")
        uri = str_url[protocol_len:]
        project_name_actual = uri.split(".")[0]
        self.assertEqual(project_name_actual, project_name)

        for v_in, v_expected, anchor in testdata_valids:
            # clean up semantic version str
            v_actual, _ = sanitize_tag(v_in)
            self.assertEqual(v_actual, v_expected)

            #    parse_ver not called yet
            sv = SemVersion()
            self.assertIsNone(sv.version_xyz())
            self.assertIsNone(sv.anchor())

            sv.parse_ver(v_in)

            self.assertEqual(anchor, sv.anchor())

            # Contains xyz, not pre or post or rc releases
            actual_url = sv.readthedocs_url(
                project_name,
                is_latest=False,
            )
            ver_xyz = sv.version_xyz()
            self.assertTrue(actual_url.endswith(ver_xyz))

            # Force get latest URL; parse_ver previously called
            str_url = sv.readthedocs_url(project_name, is_latest=True)
            self.assertTrue(str_url.endswith("latest"))

    def test_version_clean(self):
        """version_clean can take kind: current, now, tag, version str"""

        # Specify version explicitly. To transition to pre or post or rc ...
        #    Sanitizes version str
        kind = "0.0.1"  # use this instead of from git
        sv = SemVersion(path=self.cwd)
        actual_ver = sv.version_clean(kind)
        self.assertEqual(actual_ver, kind)
        self.assertEqual(actual_ver, sv.__version__)

        # invalid explicit version str
        # nowhere boys (aussie series vs some ancient power ranger like cartoon)
        kind = "five element spirit or nature?"
        with self.assertRaises(ValueError):
            sanitize_tag(kind)

        with self.assertRaises(ValueError):
            sv.version_clean(kind)

        """tag --> 0.0.1
        Avoid changing _version.py"""
        kind = "tag"
        expected = "0.0.1"
        #    Avoid changing _version.py
        with (
            patch(f"{g_app_name}.version_semantic._tag_version", return_value="0.0.1"),
        ):
            actual = sv.version_clean(kind)
            self.assertEqual(actual, expected)

        """tag. No tagged, so get current. Current found.
        Avoid changing _version.py"""
        #    Avoid changing _version.py
        with (
            patch(f"{g_app_name}.version_semantic._tag_version", return_value=None),
            patch(
                f"{g_app_name}.version_semantic._current_version", return_value=expected
            ),
        ):
            actual_ver = sv.version_clean(kind)
            self.assertEqual(actual, expected)

        """tag. Both return None; no commits at all.
        Avoid changing _version.py
        """
        kind = "tag"
        sane_fallback = "0.0.1"
        with (
            patch(f"{g_app_name}.version_semantic._tag_version", return_value=None),
            patch(
                f"{g_app_name}.version_semantic._current_version",
                return_value=sane_fallback,
            ),
        ):
            actual_ver = sv.version_clean(kind)
            self.assertEqual(actual_ver, sane_fallback)

        # current aliases. git uninitialized --> sane fallback "0.0.1"
        kinds = (
            "current",
            "now",
        )
        for kind in kinds:
            #    Avoid changing _version.py
            with (
                patch(
                    f"{g_app_name}.version_semantic._current_version",
                    return_value=sane_fallback,
                ),
            ):
                actual_ver = sv.version_clean(kind)
                self.assertEqual(actual_ver, sane_fallback)

    def test_get_app_name(self):
        """Grabs app project basename from git"""
        expected = g_app_name
        actual = _get_app_name(path=self.cwd)
        self.assertEqual(actual, expected)

        """Unlikely to ever occur. Bigger risk would be if the package
        base name is not the same as g_app_name"""
        cmd = []
        with patch(
            "subprocess.run",
            side_effect=subprocess.CalledProcessError(128, cmd),
        ):
            actual = _get_app_name(path=self.cwd)
            self.assertIsNone(actual)

        # returns nothing. Not a possibility but hardened against this
        with patch(
            "subprocess.run",
            return_value=subprocess.CompletedProcess(
                cmd,
                returncode=0,
                stdout="",
            ),
        ):
            actual = _get_app_name(path=self.cwd)
            self.assertIsNone(actual)

    def test_sanitize_kind(self):
        """Can provide a version str or get latest tag or current dev version"""
        to_tags = (
            None,
            "tag",
            (1.2345,),  # accidental tuple, but unsupported type
            1.2345,  # unsupported type
        )
        for kind in to_tags:
            actual = SemVersion.sanitize_kind(kind)
            self.assertEqual(actual, "tag")

        sames = (
            "0.0.1",
            "current",
        )
        for kind in sames:
            actual = SemVersion.sanitize_kind(kind)
            self.assertEqual(actual, kind)

        # accidental tuple
        kind = ("0.0.1",)
        actual = SemVersion.sanitize_kind(kind)
        self.assertEqual(actual, "0.0.1")

        # now is an alias of current
        kind = "now"
        actual = SemVersion.sanitize_kind(kind)
        self.assertEqual(actual, "current")

    def test_releaselevel(self):
        """releaselevel setter/getter. Ensures long form"""
        sv = SemVersion()
        for short_long, long in testdata_releaselevel:
            sv.releaselevel = short_long
            self.assertEqual(sv.releaselevel, long)

    def test_as_tuple(self):
        # prepare
        #    with dev
        #    with local
        testdata_ = (
            (
                "v0.1.1.dev0+g4b33a80.d20240129",
                (0, 1, 1, "dev0", "g4b33a80.d20240129"),
            ),
            (
                "v0.1.1.dev0",
                (0, 1, 1, "dev0"),
            ),
            (
                "v0.1.1",
                (0, 1, 1),
            ),
            ("v0.1.1.post0+g4b33a80.d20240129", (0, 1, 1, "g4b33a80.d20240129")),
        )
        for ver, t_ver_expected in testdata_:
            t_ver_actual = SemVersion.as_tuple(ver)
            self.assertEqual(t_ver_actual, t_ver_expected)

        testdata_ = (
            (
                "0.1.1.candidate1dev1+g4b33a80.d20240129",
                "0.1.1rc.dev0",
                "g4b33a80.d20240129",
            ),
        )
        for ver_orig, ver_, local_ in testdata_:
            t_ver = SemVersion.as_tuple(ver_orig)
            ver_actual = t_ver[0]
            self.assertEqual(ver_actual, ver_orig)


if __name__ == "__main__":  # pragma: no cover
    """
    w/o coverage

    .. code-block:: shell

       python -m unittest tests.test_versioning_unittest --locals

       python -m unittest tests.test_versioning_unittest \
       -k PackageVersioning.test_get_version --locals --buffer

       python -m unittest tests.test_versioning_unittest \
       -k PackageVersioning.test_sanitize_tag --locals --buffer

    W/ coverage

    .. code-block:: shell

       coverage run --data-file=".coverage-combine-31" \
       -m unittest discover -t. -s tests -p "test_versioning_unittest*.py" --locals

       coverage report --data-file=".coverage-combine-31" --no-skip-covered \
       --include="**/drain_swamp/constants*"

       coverage report --data-file=".coverage-combine-31" --no-skip-covered \
       --include="**/drain_swamp/*_version*"

    """
    unittest.main(tb_locals=True)



"""
.. moduleauthor:: Dave Faulkmore <https://mastodon.social/@msftcangoblowme>

..

Unittest for build plugins manager

.. code-block:: shell

   pytest --showlocals --log-level INFO tests/test_build_plugin_manager.py
   pytest --showlocals --cov="drain_swamp" --cov-report=term-missing tests/test_build_plugin_manager.py

"""

import pytest
from pluggy import PluginManager

import drain_swamp.monkey.plugins as package_plugins
from drain_swamp.monkey.hooks.manager import (
    get_plugin_manager,
    lazy_dotted_path,
    lazy_package,
)


def test_get_plugin_manager():
    # pytest --showlocals --log-level INFO -k "test_get_plugin_manager" tests
    # positional arg unsupported type --> TypeError
    invalids = (
        None,
        1.2,
    )
    for invalid in invalids:
        with pytest.raises(TypeError):
            get_plugin_manager(invalid)

    #
    pm = get_plugin_manager(
        package_plugins,
        namespace=None,
    )
    assert isinstance(pm, PluginManager)
    del pm
    pm = get_plugin_manager(
        package_plugins,
        specs_dotted_path=None,
    )
    assert isinstance(pm, PluginManager)
    del pm
    pm = get_plugin_manager(
        package_plugins,
        entrypoint_plugins=None,
    )
    assert isinstance(pm, PluginManager)


def test_lazy_package():
    # pytest --showlocals --log-level INFO -k "test_lazy_package" tests
    # unsupported types
    with pytest.raises(TypeError):
        lazy_package(None)
    with pytest.raises(TypeError):
        lazy_dotted_path(None)



import re
from collections.abc import Sequence
from enum import (
    Enum,
    auto,
)
from pathlib import Path
from typing import (
    Any,
    ClassVar,
)

__all__ = (
    "Snip",
    "ReplaceResult",
)

class ReplaceResult(Enum):
    VALIDATE_FAIL = auto()
    NO_MATCH = auto()
    REPLACED = auto()
    NO_CHANGE = auto()
    def __eq__(self, other: object) -> bool: ...

def check_matching_tag_count(
    contents: str,
    token_start: str | None = None,
    token_end: str | None = None,
) -> bool: ...
def check_not_nested_or_out_of_order(
    contents: str,
    token_start: str | None = None,
    token_end: str | None = None,
) -> bool: ...
def sanitize_id(id_: Any | None = "") -> str: ...

class Snip:
    TOKEN_START: ClassVar[str]
    TOKEN_END: ClassVar[str]
    PATTERN_W_ID: re.Pattern[str]

    def __init__(
        self,
        fname: str | Path,
        is_quiet: bool | None = False,
    ) -> None: ...
    @property
    def path_file(self) -> Path: ...
    @path_file.setter
    def path_file(self, val: Any) -> None: ...
    def is_file_ok(self) -> bool: ...
    @property
    def is_quiet(self) -> bool: ...
    @is_quiet.setter
    def is_quiet(self, val: Any) -> None: ...
    @property
    def is_infer(self) -> bool: ...
    @property
    def snippets(self) -> Sequence[tuple[str, str]] | ReplaceResult: ...
    def print(self) -> Sequence[tuple[str, str]] | ReplaceResult: ...
    def get_file(self) -> str: ...
    def validate(self) -> bool: ...
    def contents(self, id_: str | None = None) -> str | ReplaceResult: ...
    def replace(self, replacement: str, id_: str | None = "") -> ReplaceResult: ...
